Package: furrr
Title: Apply Mapping Functions in Parallel using Futures
Version: 0.4.0
Authors@R: c(
    person("Davis", "Vaughan", , "davis@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-4777-038X")),
    person("Henrik", "Bengtsson", , "henrikb@braju.com", role = "aut",
           comment = c(ORCID = "0000-0002-7579-5165")),
    person("Matt", "Dancho", , "mdancho@business-science.io", role = "aut"),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: Implementations of the family of map() functions from 'purrr'
    that can be resolved using any 'future'-supported backend, e.g.
    parallel on the local machine or distributed on a compute cluster.
License: MIT + file LICENSE
URL: https://github.com/futureverse/furrr,
        https://furrr.futureverse.org/
BugReports: https://github.com/futureverse/furrr/issues
Depends: future (>= 1.70.0), R (>= 4.1.0)
Imports: globals (>= 0.19.1), purrr (>= 1.2.1), rlang (>= 1.1.7), vctrs
        (>= 0.7.0)
Suggests: carrier, covr, dplyr (>= 1.1.4), knitr, parallelly (>=
        1.46.1), testthat (>= 3.3.2), tidyselect
Config/Needs/website: progressr
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-03-30 17:08:19 UTC; davis
Author: Davis Vaughan [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-4777-038X>),
  Henrik Bengtsson [aut] (ORCID: <https://orcid.org/0000-0002-7579-5165>),
  Matt Dancho [aut],
  Posit Software, PBC [cph, fnd]
Maintainer: Davis Vaughan <davis@posit.co>
Repository: CRAN
Date/Publication: 2026-03-31 14:50:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(future_modify,character)
S3method(future_modify,default)
S3method(future_modify,double)
S3method(future_modify,integer)
S3method(future_modify,logical)
S3method(future_modify,pairlist)
S3method(future_modify_at,character)
S3method(future_modify_at,default)
S3method(future_modify_at,double)
S3method(future_modify_at,integer)
S3method(future_modify_at,logical)
S3method(future_modify_if,character)
S3method(future_modify_if,default)
S3method(future_modify_if,double)
S3method(future_modify_if,integer)
S3method(future_modify_if,logical)
S3method(print,furrr_options)
export(furrr_options)
export(future_imap)
export(future_imap_chr)
export(future_imap_dbl)
export(future_imap_dfc)
export(future_imap_dfr)
export(future_imap_int)
export(future_imap_lgl)
export(future_imap_vec)
export(future_iwalk)
export(future_map)
export(future_map2)
export(future_map2_chr)
export(future_map2_dbl)
export(future_map2_dfc)
export(future_map2_dfr)
export(future_map2_int)
export(future_map2_lgl)
export(future_map2_vec)
export(future_map_at)
export(future_map_chr)
export(future_map_dbl)
export(future_map_dfc)
export(future_map_dfr)
export(future_map_if)
export(future_map_int)
export(future_map_lgl)
export(future_map_vec)
export(future_modify)
export(future_modify_at)
export(future_modify_if)
export(future_pmap)
export(future_pmap_chr)
export(future_pmap_dbl)
export(future_pmap_dfc)
export(future_pmap_dfr)
export(future_pmap_int)
export(future_pmap_lgl)
export(future_pmap_vec)
export(future_pwalk)
export(future_walk)
export(future_walk2)
import(rlang)
importFrom(future,future)
importFrom(parallel,nextRNGStream)
importFrom(parallel,nextRNGSubStream)
importFrom(purrr,map)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# furrr <a href='https://furrr.futureverse.org/'><img src='man/figures/logo.png' align="right" height="139" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/furrr)](https://cran.r-project.org/package=furrr)
[![R-CMD-check](https://github.com/futureverse/furrr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/futureverse/furrr/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/futureverse/furrr/graph/badge.svg)](https://app.codecov.io/gh/futureverse/furrr)
<!-- badges: end -->

## Overview

The goal of furrr is to combine purrr’s family of mapping functions with
future’s parallel processing capabilities. The result is near drop in
replacements for purrr functions such as `map()` and `map2_dbl()`, which
can be replaced with their furrr equivalents of `future_map()` and
`future_map2_dbl()` to map in parallel.

The code draws heavily from the implementations of purrr and
future.apply and this package would not be possible without either of
them.

## What has been implemented?

Every variant of the following functions has been implemented:

- `map()`
- `map2()`
- `pmap()`
- `walk()`
- `imap()`
- `modify()`

This includes atomic variants like `map_dbl()` through
`future_map_dbl()` and predicate variants like `map_at()` through
`future_map_at()`.

## Installation

You can install the released version of furrr from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("furrr")
```

And the development version from [GitHub](https://github.com/) with:

``` r
# install.packages("remotes")
remotes::install_github("futureverse/furrr")
```

## Learning

The easiest way to learn about furrr is to browse [the
website](https://furrr.futureverse.org/). In particular, the [function
reference](https://furrr.futureverse.org/reference/index.html) page can
be useful to get a general overview of the functions in the package, and
the following vignettes are deep dives into various parts of furrr:

- [Common
  gotchas](https://furrr.futureverse.org/articles/articles/gotchas.html)

- [Learn how furrr “chunks” your
  input](https://furrr.futureverse.org/articles/articles/chunking.html)

- [carrier - An alternative to automatic globals
  detection](https://furrr.futureverse.org/articles/articles/carrier.html)

- [Progress notifications with
  progressr](https://furrr.futureverse.org/articles/articles/progress.html)

- [Using furrr with
  connections](https://furrr.futureverse.org/articles/articles/remote-connections.html)

## Example

furrr has been designed to function as identically to purrr as possible,
so that you can immediately have familiarity with it.

``` r
library(furrr)
library(purrr)

map(c("hello", "world"), ~.x)
#> [[1]]
#> [1] "hello"
#> 
#> [[2]]
#> [1] "world"

future_map(c("hello", "world"), ~.x)
#> [[1]]
#> [1] "hello"
#> 
#> [[2]]
#> [1] "world"
```

The default backend for future (and through it, furrr) is a sequential
one. This means that the above code will run out of the box, but it will
*not* be in parallel. The design of future makes it incredibly easy to
change this so that your code will run in parallel.

``` r
# Set a "plan" for how the code should run.
plan(multisession, workers = 2)

# This does run in parallel!
future_map(c("hello", "world"), ~.x)
#> [[1]]
#> [1] "hello"
#> 
#> [[2]]
#> [1] "world"
```

If you are still skeptical, here is some proof that we are running in
parallel.

``` r
library(tictoc)

# This should take 6 seconds in total running sequentially
plan(sequential)

tic()
nothingness <- future_map(c(2, 2, 2), ~ Sys.sleep(.x))
toc()
#> 6.08 sec elapsed
```

``` r
# This should take ~2 seconds running in parallel, with a little overhead
# in `future_map()` from sending data to the workers. There is generally also
# a one time cost from `plan(multisession)` setting up the workers.
plan(multisession, workers = 3)

tic()
nothingness <- future_map(c(2, 2, 2), ~ Sys.sleep(.x))
toc()
#> 2.212 sec elapsed
```

## Data transfer

It’s important to remember that data has to be passed back and forth
between the workers. This means that whatever performance gain you might
have gotten from your parallelization can be crushed by moving large
amounts of data around. For example, if you are moving large data frames
to the workers, running models in parallel, and returning large model
objects back, the shuffling of data can take a large chunk of that time.
Rather than returning the entire model object, you might consider only
returning a performance metric, or smaller specific pieces of that model
that you are most interested in.

This performance drop can especially be prominent if using
`future_pmap()` to iterate over rows and return large objects at each
iteration.

#' Apply a function to each element of a vector via futures
#'
#' These functions work the same as [purrr::map()] and its variants, but
#' allow you to map in parallel.
#'
#' @param .x A list or atomic vector.
#'
#' @param .f A function, specified in one of the following ways:
#'
#'   * A named function, e.g. `mean`.
#'   * An anonymous function, e.g. `\(x) x + 1` or `function(x) x + 1`.
#'   * A formula, e.g. `~ .x + 1`. Use `.x` to refer to the first
#'     argument. No longer recommended.
#'   * A string, integer, or list, e.g. `"idx"`, `1`, or `list("idx", 1)` which
#'     are shorthand for `\(x) pluck(x, "idx")`, `\(x) pluck(x, 1)`, and
#'     `\(x) pluck(x, "idx", 1)` respectively. Optionally supply `.default` to
#'     set a default value if the indexed element is `NULL` or does not exist.
#'
#' @param ... Additional arguments passed on to the mapped function.
#'
#'   We now generally recommend against using `...` to pass additional
#'   (constant) arguments to `.f`. Instead use a shorthand anonymous function:
#'
#'   ```R
#'   # Instead of
#'   x |> future_map(f, 1, 2, collapse = ",")
#'   # do:
#'   x |> future_map(\(x) f(x, 1, 2, collapse = ","))
#'   ```
#'
#'   This makes it easier to understand which arguments belong to which
#'   function and will tend to yield better error messages.
#'
#' @param .id Either a string or `NULL`. If a string, the output will contain
#'   a variable with that name, storing either the name (if `.x` is named) or
#'   the index (if `.x` is unnamed) of the input. If `NULL`, the default, no
#'   variable will be created.
#'
#'   Only applies to `_dfr` variant.
#'
#' @param .env_globals The environment to look for globals required by `.x` and
#'   `...`. Globals required by `.f` are looked up in the function environment
#'   of `.f`.
#'
#' @param .options The `future` specific options to use with the workers. This
#'   must be the result from a call to [furrr_options()].
#'
#' @param .progress A single logical. Should a progress bar be displayed?
#'   Only works with multisession, multicore, and multiprocess futures. Note
#'   that if a multicore/multisession future falls back to sequential, then
#'   a progress bar will not be displayed.
#'
#'   __Warning:__ The `.progress` argument will be deprecated and removed
#'   in a future version of furrr in favor of using the more robust
#'   [progressr](https://CRAN.R-project.org/package=progressr)
#'   package.
#'
#' @param .ptype If `NULL`, the default, the output type is the common type of
#'   the elements of the result. Otherwise, supply a "prototype" giving the
#'   desired type of output.
#'
#' @return All functions return a vector the same length as `.x`.
#'
#' - [future_map()] returns a list
#'
#' - [future_map_lgl()] a logical vector
#'
#' - [future_map_int()] an integer vector
#'
#' - [future_map_dbl()] a double vector
#'
#' - [future_map_chr()] a character vector
#'
#' The output of `.f` will be automatically typed upwards, e.g. logical ->
#' integer -> double -> character.
#'
#' @export
#' @examples
#' \donttest{plan(multisession, workers = 2)}
#'
#' 1:10 |>
#'   future_map(rnorm, n = 10, .options = furrr_options(seed = 123)) |>
#'   future_map_dbl(mean)
#'
#' # If each element of the output is a data frame, use
#' # `future_map_dfr()` to row-bind them together:
#' mtcars |>
#'   split(mtcars$cyl) |>
#'   future_map(~ lm(mpg ~ wt, data = .x)) |>
#'   future_map_dfr(~ as.data.frame(t(as.matrix(coef(.)))))
#'
#'
#' # You can be explicit about what gets exported to the workers.
#' # To see this, use multisession (not multicore as the forked workers
#' # still have access to this environment)
#' \donttest{plan(multisession)}
#' x <- 1
#' y <- 2
#'
#' # This will fail, y is not exported (no black magic occurs)
#' try(future_map(1, ~y, .options = furrr_options(globals = "x")))
#'
#' # y is exported
#' future_map(1, ~y, .options = furrr_options(globals = "y"))
#'
#' \dontshow{
#' # Close open connections for R CMD Check
#' if (!inherits(plan(), "sequential")) plan(sequential)
#' }
future_map <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map_template(
    x = .x,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = "map",
    env_globals = .env_globals
  )
}

#' @rdname future_map
#' @export
future_map_chr <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map_template(
    x = .x,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "character",
    purrr_fn_name = "map_chr",
    env_globals = .env_globals
  )
}

#' @rdname future_map
#' @export
future_map_dbl <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map_template(
    x = .x,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "double",
    purrr_fn_name = "map_dbl",
    env_globals = .env_globals
  )
}

#' @rdname future_map
#' @export
future_map_int <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map_template(
    x = .x,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "integer",
    purrr_fn_name = "map_int",
    env_globals = .env_globals
  )
}

#' @rdname future_map
#' @export
future_map_lgl <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map_template(
    x = .x,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "logical",
    purrr_fn_name = "map_lgl",
    env_globals = .env_globals
  )
}

#' @rdname future_map
#' @export
future_map_vec <- function(
  .x,
  .f,
  ...,
  .ptype = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  out <- future_map(
    .x = .x,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  simplify_impl(
    out,
    ptype = .ptype,
    error_arg = "<output>",
    error_call = current_env()
  )
}

#' @rdname future_map
#' @export
future_map_dfr <- function(
  .x,
  .f,
  ...,
  .id = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  if (!rlang::is_installed("dplyr")) {
    rlang::abort("`future_map_dfr()` requires dplyr")
  }

  res <- future_map(
    .x = .x,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  dplyr::bind_rows(res, .id = .id)
}

#' @rdname future_map
#' @export
future_map_dfc <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  if (!rlang::is_installed("dplyr")) {
    rlang::abort("`future_map_dfc()` requires dplyr")
  }

  res <- future_map(
    .x = .x,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  dplyr::bind_cols(res)
}

# ------------------------------------------------------------------------------

#' Apply a function to each element of a vector conditionally via futures
#'
#' These functions work the same as [purrr::map_if()] and
#' [purrr::map_at()], but allow you to run them in parallel.
#'
#' @inheritParams future_map
#'
#' @param .p A single predicate function, a formula describing such a
#'   predicate function, or a logical vector of the same length as `.x`.
#'   Alternatively, if the elements of `.x` are themselves lists of
#'   objects, a string indicating the name of a logical element in the
#'   inner lists. Only those elements where `.p` evaluates to
#'   `TRUE` will be modified.
#'
#' @param .else A function applied to elements of `.x` for which `.p`
#' returns `FALSE`.
#'
#' @param .at A logical, integer, or character vector giving the elements
#'   to select. Alternatively, a function that takes a vector of names,
#'   and returns a logical, integer, or character vector of elements to select.
#'
#'   `r lifecycle::badge("deprecated")`: if the tidyselect package is
#'   installed, you can use `vars()` and tidyselect helpers to select
#'   elements.
#'
#' @return Both functions return a list the same length as `.x` with the
#' elements conditionally transformed.
#'
#' @export
#' @examples
#' \donttest{plan(multisession, workers = 2)}
#'
#' # Modify the even elements
#' future_map_if(1:5, ~.x %% 2 == 0L, ~ -1)
#'
#' future_map_at(1:5, c(1, 5), ~ -1)
#' \dontshow{
#' # Close open connections for R CMD Check
#' if (!inherits(plan(), "sequential")) plan(sequential)
#' }
future_map_if <- function(
  .x,
  .p,
  .f,
  ...,
  .else = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  sel <- probe(.x, .p)
  out <- vector("list", length = length(.x))

  out[sel] <- future_map(
    .x = .x[sel],
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  if (is_null(.else)) {
    out[!sel] <- .x[!sel]
  } else {
    out[!sel] <- future_map(
      .x = .x[!sel],
      .f = .else,
      ...,
      .options = .options,
      .env_globals = .env_globals,
      .progress = FALSE
    )
  }

  set_names(out, names(.x))
}

#' @rdname future_map_if
#' @export
future_map_at <- function(
  .x,
  .at,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  sel <- inv_which(.x, .at)
  out <- vector("list", length = length(.x))

  out[sel] <- future_map(
    .x = .x[sel],
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  out[!sel] <- .x[!sel]

  set_names(out, names(.x))
}

# ------------------------------------------------------------------------------
# furrr_options(globals =)

test_that("can export globals using a named list", {
  plan(multisession, workers = 2)
  on.exit(plan(sequential), add = TRUE)

  opts <- furrr_options(globals = list(y = 1))

  fn <- function(x) {
    exists("y")
  }

  expect_identical(future_map_lgl(1:2, fn, .options = opts), c(TRUE, TRUE))
})

test_that("can detect globals from the caller environment (HenrikBengtsson/future.apply#62)", {
  plan(multisession, workers = 2)
  on.exit(plan(sequential), add = TRUE)

  fn <- function(x) {
    exists("y")
  }

  wrapper <- function(options = furrr_options()) {
    y <- 1
    future_map_lgl(1, fn, .options = options)
  }

  expect_identical(wrapper(), FALSE)

  options <- furrr_options(globals = "y")
  expect_identical(wrapper(options), TRUE)
})

test_that("can export globals with sequential futures", {
  plan(sequential)

  fn <- function(x) {
    exists("y")
  }
  environment(fn) <- .GlobalEnv

  # With named list
  opts <- furrr_options(globals = list(y = 1))
  expect_identical(future_map_lgl(1:2, fn, .options = opts), c(TRUE, TRUE))

  # With character lookup in caller env
  wrapper <- function(fn) {
    y <- 1
    future_map_lgl(1, fn, .options = furrr_options(globals = "y"))
  }
  environment(wrapper) <- .GlobalEnv

  expect_identical(wrapper(fn), TRUE)
})

test_that("validates `globals`", {
  expect_error(furrr_options(globals = 1))
  expect_error(furrr_options(globals = c(TRUE, TRUE)))
  expect_error(furrr_options(globals = list(a = 1, 2)))
})

# ------------------------------------------------------------------------------
# furrr_options(packages =)

test_that("can selectively export packages on multisession", {
  plan(multisession, workers = 2)
  on.exit(plan(sequential), add = TRUE)

  opts <- furrr_options(packages = "dplyr")

  # TODO: Reenable this test after future issue is fixed
  # https://github.com/futureverse/future/issues/820
  # https://github.com/futureverse/furrr/issues/307
  # expect_error(
  #   future_map(1:2, ~ tibble(x = .x))
  # )

  expect_identical(
    future_map(1:2, ~ tibble(x = .x), .options = opts),
    list(dplyr::tibble(x = 1L), dplyr::tibble(x = 2L))
  )
})

test_that("validates `packages`", {
  expect_error(furrr_options(packages = 1))
})

# ------------------------------------------------------------------------------
# furrr_options(stdout =)

furrr_test_that("can use `stdout = FALSE`", {
  opts <- furrr_options(stdout = FALSE)

  fn <- function(x) {
    con <- stdout()
    write("hello", con)
    x
  }

  expect_silent(future_map(1:5, fn, .options = opts))
})

test_that("can use `stdout = NA` to not intercept at all", {
  opts <- furrr_options(stdout = NA)

  fn <- function(x) {
    con <- stdout()
    write("hello", con)
    x
  }

  # Output is likely shown for sequential backends
  plan(sequential)
  expect_snapshot(future_map(1:2, fn, .options = opts))

  # Output is likely not passed back from the other R session
  # (because it isn't handled or captured at all)
  plan(multisession, workers = supported_max_cores("multisession"))
  on.exit(plan(sequential), add = TRUE)
  expect_snapshot(future_map(1:2, fn, .options = opts))
})

test_that("validates `stdout`", {
  expect_snapshot({
    (expect_error(furrr_options(stdout = "x")))
    (expect_error(furrr_options(stdout = c(TRUE, TRUE))))
  })
})

# ------------------------------------------------------------------------------
# furrr_options(conditions =)

furrr_test_that("can capture no conditions", {
  fn <- function(x) {
    warning("hello")
    x
  }

  opts <- furrr_options(conditions = character())
  expect_warning(future_map(1:5, fn, .options = opts), NA)
})

test_that("can avoid handling conditions altogether", {
  fn <- function(x) {
    warning("hello")
    x
  }

  opts <- furrr_options(conditions = NULL)

  # Warning is likely shown for sequential backends
  plan(sequential)
  expect_snapshot(future_map(1:5, fn, .options = opts))

  # Warning is likely not passed back from the other R session
  # (because it isn't handled or captured at all)
  plan(multisession, workers = supported_max_cores("multisession"))
  on.exit(plan(sequential), add = TRUE)
  expect_snapshot(future_map(1:5, fn, .options = opts))
})

furrr_test_that("can selectively avoid conditions", {
  fn <- function(x) {
    rlang::warn("classed warning", class = "ignore_me")
    rlang::warn("unclassed warning", class = "dont_ignore_me")
    x
  }

  # Both warnings are shown
  expect_warning(
    expect_warning(future_map(1, fn), class = "ignore_me"),
    class = "dont_ignore_me"
  )

  # Only dont_ignore_me is shown
  opts <- furrr_options(
    conditions = structure("condition", exclude = "ignore_me")
  )
  expect_warning(future_map(1, fn, .options = opts), class = "dont_ignore_me")
})

test_that("validates `conditions`", {
  expect_snapshot({
    (expect_error(furrr_options(conditions = 1)))
  })
})

# ------------------------------------------------------------------------------
# furrr_options(seed =)

furrr_test_that("can use integer `seed` of size 1", {
  old_seed <- get_random_seed()

  opts <- furrr_options(seed = 123)

  x <- future_map(5, runif, .options = opts)

  expect_false(identical(old_seed, get_random_seed()))

  y <- future_map(5, runif, .options = opts)

  expect_identical(x, y)
})

furrr_test_that("can use integer `seed` of size 7", {
  old_rngkind <- RNGkind("L'Ecuyer-CMRG")
  lecuyer_seed <- get_random_seed()

  # Reset
  RNGkind(old_rngkind[[1]])
  old_seed <- get_random_seed()

  opts <- furrr_options(seed = lecuyer_seed)

  x <- future_map(5, runif, .options = opts)

  expect_false(identical(old_seed, get_random_seed()))

  y <- future_map(5, runif, .options = opts)

  expect_identical(x, y)
})

furrr_test_that("`TRUE` advances `.Random.seed` but `FALSE` doesn't", {
  old_seed <- get_random_seed()

  opts <- furrr_options(seed = FALSE)
  future_map(1:5, ~.x, .options = opts)

  expect_identical(old_seed, get_random_seed())

  opts <- furrr_options(seed = TRUE)
  future_map(1:5, ~.x, .options = opts)

  expect_false(identical(old_seed, get_random_seed()))
})

furrr_test_that("`NULL` seed doesn't advance `.Random.seed`", {
  old_seed <- get_random_seed()

  opts <- furrr_options(seed = NULL)
  future_map(1:5, ~.x, .options = opts)

  expect_identical(old_seed, get_random_seed())
})

test_that("must use valid L'Ecuyer-CMRG seed", {
  opts <- furrr_options(seed = 1:7)
  expect_error(future_map(1:5, ~.x, .options = opts), "must be")
})

test_that("validates `seed`", {
  expect_error(furrr_options(seed = "x"))
  expect_error(furrr_options(seed = c(1, 2)))
  expect_error(furrr_options(seed = c(TRUE, FALSE)))
  expect_error(furrr_options(seed = 1.5))
  expect_error(furrr_options(seed = NA))
})

# ------------------------------------------------------------------------------
# furrr_options(scheduling =)

test_that("can specify `scheduling`", {
  x <- furrr_options(scheduling = TRUE)
  expect_identical(x$scheduling, TRUE)

  x <- furrr_options(scheduling = FALSE)
  expect_identical(x$scheduling, FALSE)

  x <- furrr_options(scheduling = 0)
  expect_identical(x$scheduling, 0L)

  x <- furrr_options(scheduling = 5)
  expect_identical(x$scheduling, 5L)

  x <- furrr_options(scheduling = Inf)
  expect_identical(x$scheduling, Inf)
})

test_that("validates `scheduling`", {
  expect_error(furrr_options(scheduling = c(1, 2)))
  expect_error(furrr_options(scheduling = c(TRUE, FALSE)))
  expect_error(furrr_options(scheduling = "x"))
  expect_error(furrr_options(scheduling = 1.5))
  expect_error(furrr_options(scheduling = NA))
  expect_error(furrr_options(scheduling = -Inf))
  expect_error(furrr_options(scheduling = lm(1 ~ 1)))
})

test_that("`scheduling` supports an `ordering` attribute (#289)", {
  # Integer `scheduling`
  x <- furrr_options(scheduling = structure(2L, ordering = "random"))
  expect_identical(x$scheduling, structure(2L, ordering = "random"))

  # Double `scheduling`
  x <- furrr_options(scheduling = structure(2, ordering = "random"))
  expect_identical(x$scheduling, structure(2L, ordering = "random"))

  # Inf `scheduling`
  x <- furrr_options(scheduling = structure(Inf, ordering = "random"))
  expect_identical(x$scheduling, structure(Inf, ordering = "random"))

  # Logical `scheduling`
  x <- furrr_options(scheduling = structure(TRUE, ordering = "random"))
  expect_identical(x$scheduling, structure(TRUE, ordering = "random"))
})

# ------------------------------------------------------------------------------
# furrr_options(chunk_size =)

test_that("can specify `chunk_size`", {
  x <- furrr_options(chunk_size = 2)
  expect_identical(x$chunk_size, 2L)

  x <- furrr_options(chunk_size = Inf)
  expect_identical(x$chunk_size, Inf)
})

test_that("validates `chunk_size`", {
  expect_error(furrr_options(chunk_size = 0))
  expect_error(furrr_options(chunk_size = c(1, 2)))
  expect_error(furrr_options(chunk_size = "x"))
  expect_error(furrr_options(chunk_size = 1.5))
  expect_error(furrr_options(chunk_size = NA))
  expect_error(furrr_options(chunk_size = -Inf))
  expect_error(furrr_options(chunk_size = lm(1 ~ 1)))
})

test_that("`chunk_size` supports an `ordering` attribute (#290)", {
  # Integer `chunk_size`
  x <- furrr_options(chunk_size = structure(2L, ordering = "random"))
  expect_identical(x$chunk_size, structure(2L, ordering = "random"))

  # Double `chunk_size`
  x <- furrr_options(chunk_size = structure(2, ordering = "random"))
  expect_identical(x$chunk_size, structure(2L, ordering = "random"))

  # Inf `chunk_size`
  x <- furrr_options(chunk_size = structure(Inf, ordering = "random"))
  expect_identical(x$chunk_size, structure(Inf, ordering = "random"))
})

# ------------------------------------------------------------------------------
# furrr_options(prefix =)

furrr_test_that("labels can be passed through without erroring", {
  opts <- furrr_options(prefix = "foo")
  expect_error(future_map(1:5, ~.x, .options = opts), NA)
})

test_that("validates `prefix`", {
  expect_error(furrr_options(prefix = 1))
  expect_error(furrr_options(prefix = c("x", "y")))
})

# ------------------------------------------------------------------------------
# validation

test_that("option object is validated", {
  expect_error(future_map(1:5, ~.x, .options = 1))
  expect_error(future_map2(1:5, 1:5, ~.x, .options = 1))
  expect_error(future_pmap(list(1:5), ~.x, .options = 1))
})

# ------------------------------------------------------------------------------
# map()

furrr_test_that("future_map() matches map() for simple cases", {
  expect_identical(
    future_map(1:3, ~.x),
    map(1:3, ~.x)
  )
})

furrr_test_that("names of `.x` are retained", {
  x <- c(a = 1, b = 2)
  expect_named(future_map(x, ~1), c("a", "b"))
})

furrr_test_that("named empty input makes named empty output", {
  x <- set_names(list(), character())
  expect_named(future_map(x, ~.x), character())
})

# ------------------------------------------------------------------------------
# atomic variants

furrr_test_that("future_map_dbl() works", {
  x <- c(1, 2, 3)

  expect_identical(
    future_map_dbl(x, ~.x),
    map_dbl(x, ~.x)
  )
})

furrr_test_that("future_map_int() works", {
  x <- c(1L, 2L, 3L)

  expect_identical(
    future_map_int(x, ~.x),
    map_int(x, ~.x)
  )
})

furrr_test_that("future_map_lgl() works", {
  x <- c(TRUE, FALSE, TRUE)

  expect_identical(
    future_map_lgl(x, ~.x),
    map_lgl(x, ~.x)
  )
})

furrr_test_that("future_map_chr() works", {
  x <- c("a", "b", "c")

  expect_identical(
    future_map_chr(x, ~.x),
    map_chr(x, ~.x)
  )
})

furrr_test_that("future_map_vec() works", {
  x <- as.Date(c("2020-01-01", "2020-01-02", "2020-01-03"))

  expect_identical(
    future_map_vec(x, ~.x),
    map_vec(x, ~.x)
  )
  expect_identical(
    future_map_vec(x, ~1, .ptype = integer()),
    map_vec(x, ~1, .ptype = integer())
  )

  expect_identical(
    future_map_vec(integer(), identity),
    map_vec(integer(), identity)
  )
  expect_identical(
    future_map_vec(set_names(integer(), character()), identity),
    map_vec(set_names(integer(), character()), identity)
  )

  # Vector error
  expect_snapshot(error = TRUE, {
    future_map_vec(1:2, ~NULL)
  })
  # Size error
  expect_snapshot(error = TRUE, {
    future_map_vec(1:2, ~ 1:2)
  })
  # Type error
  expect_snapshot(error = TRUE, {
    future_map_vec(1:2, ~ if (.x == 1L) 1 else "x")
  })
})

furrr_test_that("names of `.x` are retained", {
  x <- c(a = 1, b = 2)
  expect_named(future_map_dbl(x, ~1), c("a", "b"))

  x <- c(a = as.Date("2020-01-01"), b = as.Date("2020-01-02"))
  expect_named(future_map_vec(x, ~1), c("a", "b"))
})

# ------------------------------------------------------------------------------
# data frame variants

furrr_test_that("future_map_dfr() works", {
  x <- c("a", "b", "c")

  expect_identical(
    future_map_dfr(x, ~ data.frame(x = .x)),
    map_dfr(x, ~ data.frame(x = .x))
  )
})

furrr_test_that("future_map_dfc() works", {
  x <- c("a", "b", "c")

  expect_identical(
    future_map_dfc(x, ~ as.data.frame(set_names(list(1), .x))),
    map_dfc(x, ~ as.data.frame(set_names(list(1), .x)))
  )
})

# ------------------------------------------------------------------------------
# size

furrr_test_that("future_map() works with size zero input", {
  expect_identical(future_map(list(), identity), list())
})

furrr_test_that("atomic variants work with size zero input", {
  expect_identical(future_map_chr(list(), identity), character())
  expect_identical(future_map_dbl(list(), identity), double())
  expect_identical(future_map_int(list(), identity), integer())
  expect_identical(future_map_lgl(list(), identity), logical())
})

furrr_test_that("generic variant works with size zero input", {
  expect_identical(
    future_map_vec(list(), identity),
    NULL
  )
  expect_identical(
    future_map_vec(list(), identity, .ptype = integer()),
    integer()
  )
})

# ------------------------------------------------------------------------------
# at / if variants

furrr_test_that("future_map_at() works", {
  x <- list("a", "b", "c")

  expect_identical(
    future_map_at(x, 2, ~3),
    map_at(x, 2, ~3)
  )
})

furrr_test_that("names of `.x` are retained", {
  x <- list(a = "a", b = "b", c = "c")
  expect_named(future_map_at(x, 2, ~3), c("a", "b", "c"))
})

furrr_test_that("future_map_if() works", {
  x <- list("a", "b", "c")

  expect_identical(
    future_map_if(x, ~ .x %in% c("a", "c"), ~3),
    map_if(x, ~ .x %in% c("a", "c"), ~3)
  )
})

furrr_test_that("names of `.x` are retained", {
  x <- list(a = "a", b = "b", c = "c")
  expect_named(future_map_if(x, ~ .x %in% c("a", "c"), ~3), c("a", "b", "c"))
})

furrr_test_that("`.else` can be used", {
  x <- list("a", "b", "c")

  expect_identical(
    future_map_if(x, ~ .x %in% c("a", "c"), ~3, .else = ~ -1),
    map_if(x, ~ .x %in% c("a", "c"), ~3, .else = ~ -1)
  )
})

# ------------------------------------------------------------------------------
# Miscellaneous

furrr_test_that("Calling `~` from within `.f` works", {
  x <- list(
    list(a = 4, b = 6),
    list(c = 5, d = 7)
  )

  expect_identical(future_map(x, ~ map(.x, ~.x)), x)
})

furrr_test_that("Calling `~` from within `.f` inside a `mutate()` works (#7, #123)", {
  x <- list(
    list(a = 4, b = 6),
    list(c = 5, d = 7)
  )

  df <- dplyr::tibble(x = x)

  expect_identical(
    dplyr::mutate(df, x = future_map(x, ~ map(.x, ~.x))),
    df
  )
})

furrr_test_that("globals in `.x` are found (#16)", {
  fn <- function(x) sum(x, na.rm = TRUE)

  x <- list(c(1, 2, NA), c(2, 3, 4))

  fns1 <- map(x, ~ purrr::partial(fn, x = .x))
  fns2 <- map(x, ~ function() fn(.x))

  expect_identical(future_map_dbl(fns1, ~ .x()), c(3, 9))
  expect_identical(future_map_dbl(fns2, ~ .x()), c(3, 9))
})

test_that("globals in `.x` are only exported to workers that use them", {
  plan(multisession, workers = 2)
  on.exit(plan(sequential), add = TRUE)

  # Use `local()` to ensure that the wrapper functions and the anonymous
  # functions created with `~` don't pick up extra globals
  my_wrapper1 <- local({
    my_mean1 <- function(x) mean(x, na.rm = TRUE)

    function(x) {
      my_mean1(x)
      exists("my_mean1")
    }
  })

  my_wrapper2 <- local({
    my_mean2 <- function(x) mean(x, na.rm = FALSE)

    function(x) {
      my_mean2(x)
      exists("my_mean1")
    }
  })

  x <- list(my_wrapper1, my_wrapper2)

  expect_identical(
    future_map_lgl(.x = x, .f = ~ .x(c(1, NA))),
    c(TRUE, FALSE)
  )
})

furrr_test_that("base package functions can be exported to workers (HenrikBengtsson/future#401)", {
  expect_identical(future_map(1:2, identity), list(1L, 2L))
})

furrr_test_that("`.f` globals are only looked up in the function env of `.f` (#153)", {
  fn <- function(x) {
    y
  }

  fn2 <- local({
    y <- -1

    function(x) {
      y
    }
  })

  wrapper <- function(f) {
    y <- 1
    future_map(1:2, f)
  }

  # TODO: Reenable this test after future issue is fixed
  # https://github.com/futureverse/future/issues/820
  # https://github.com/futureverse/furrr/issues/307
  # expect_error(wrapper(fn), "'y' not found")

  expect_identical(wrapper(fn2), list(-1, -1))
})

furrr_test_that("`...` globals/packages are found", {
  # We set the function environments to the global environment to ensure
  # that they aren't set to something else while `test()` is running

  fn <- function(x, fn_arg) {
    fn_arg()
  }
  environment(fn) <- .GlobalEnv

  fn_arg_env <- new_environment(list(x = 1), parent = .GlobalEnv)

  # This function is passed through `...`
  fn_arg <- function() {
    x
  }
  environment(fn_arg) <- fn_arg_env

  expect_identical(
    future_map(1:2, fn, fn_arg = fn_arg),
    list(1, 1)
  )
})

furrr_test_that("errors don't report purrr's indices (#250)", {
  fail_on_five <- function(x) {
    if (x == 5) {
      stop("Failure!")
    }
    1L
  }

  x <- 1:20

  expect_snapshot(error = TRUE, {
    future_map(x, fail_on_five)
  })
})

#' @rdname future_map
#' @export
future_walk <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map_template(
    x = .x,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = "walk",
    env_globals = .env_globals
  )

  invisible(.x)
}

#' @rdname future_map2
#' @export
future_walk2 <- function(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map2_template(
    x = .x,
    y = .y,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = "walk2",
    env_globals = .env_globals
  )

  invisible(.x)
}

#' @rdname future_map2
#' @export
future_pwalk <- function(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_pmap_template(
    l = .l,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = "pwalk",
    env_globals = .env_globals
  )

  invisible(.l)
}

#' @rdname future_imap
#' @export
future_iwalk <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_walk2(
    .x = .x,
    .y = vec_index(.x),
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )
}

furrr_test_that("walk functions work", {
  x <- 1:5
  out <- future_walk(x, ~"hello")
  expect_identical(out, x)

  y <- 6:10
  out <- future_walk2(x, y, ~"hello")
  expect_identical(out, x)

  l <- list(x, y)
  out <- future_pwalk(list(x, y), ~"hello")
  expect_identical(out, l)

  out <- future_iwalk(x, ~"hello")
  expect_identical(out, x)
})

furrr_test_that("`future_walk()` - template returns `NULL`s (#205)", {
  .x <- 1:2
  .f <- identity
  .options <- furrr_options()
  .progress <- FALSE
  .env_globals <- environment()

  # `map()` template returns `.f` output
  purrr_fn_name <- "map"
  out <- furrr_map_template(
    x = .x,
    fn = .f,
    dots = list(),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = purrr_fn_name,
    env_globals = .env_globals
  )
  expect_identical(out, list(1L, 2L))

  # `walk()` template returns `NULL`s
  purrr_fn_name <- "walk"
  out <- furrr_map_template(
    x = .x,
    fn = .f,
    dots = list(),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = purrr_fn_name,
    env_globals = .env_globals
  )
  expect_identical(out, list(NULL, NULL))
})

furrr_test_that("`future_walk2()` - template returns `NULL`s (#205)", {
  .x <- 1:2
  .y <- 3:4
  .f <- function(x, y) c(x, y)
  .options <- furrr_options()
  .progress <- FALSE
  .env_globals <- environment()

  # `map2()` template returns `.f` output
  purrr_fn_name <- "map2"
  out <- furrr_map2_template(
    x = .x,
    y = .y,
    fn = .f,
    dots = list(),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = purrr_fn_name,
    env_globals = .env_globals
  )
  expect_identical(out, list(c(1L, 3L), c(2L, 4L)))

  # `walk2()` template returns `NULL`s
  purrr_fn_name <- "walk2"
  out <- furrr_map2_template(
    x = .x,
    y = .y,
    fn = .f,
    dots = list(),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = purrr_fn_name,
    env_globals = .env_globals
  )
  expect_identical(out, list(NULL, NULL))
})

furrr_test_that("`future_pwalk()` - template returns `NULL`s (#205)", {
  .l <- list(1:2, 3:4, 5:6)
  .f <- function(x, y, z) c(x, y, z)
  .options <- furrr_options()
  .progress <- FALSE
  .env_globals <- environment()

  # `pmap()` template returns `.f` output
  purrr_fn_name <- "pmap"
  out <- furrr_pmap_template(
    l = .l,
    fn = .f,
    dots = list(),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = purrr_fn_name,
    env_globals = .env_globals
  )
  expect_identical(out, list(c(1L, 3L, 5L), c(2L, 4L, 6L)))

  # `pwalk()` template returns `NULL`s
  purrr_fn_name <- "pwalk"
  out <- furrr_pmap_template(
    l = .l,
    fn = .f,
    dots = list(),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = purrr_fn_name,
    env_globals = .env_globals
  )
  expect_identical(out, list(NULL, NULL))
})

#' Options to fine tune furrr
#'
#' `furrr_options()` returns an object that can be supplied as the `.options`
#' argument for furrr functions, such as [future_map()]. The options are either
#' used by furrr directly, or are passed on to [future::future()].
#'
#' @param ... These dots are reserved for future extensibility and must
#'   be empty.
#'
#' @param stdout A logical.
#'
#'   - If `TRUE`, standard output of the underlying futures is relayed as soon
#'   as possible.
#'
#'   - If `FALSE`, output is silenced by sinking it to the null device.
#'
#' @param conditions A character string of conditions classes to be relayed.
#'   The default is to relay all conditions, including messages and warnings.
#'   Errors are always relayed. To not relay any conditions (besides errors),
#'   use `conditions = character()`. To selectively ignore specific classes,
#'   use `conditions = structure("condition", exclude = "message")`.
#'
#' @param globals A logical, a character vector, a named list, or `NULL` for
#'   controlling how globals are handled. For details, see the
#'   `Global variables` section below.
#'
#' @param packages A character vector, or `NULL`. If supplied, this specifies
#'   packages that are guaranteed to be attached in the R environment where the
#'   future is evaluated.
#'
#' @param seed A logical, an integer of length `1` or `7`, a list of
#'   `length(.x)` with pre-generated random seeds, or `NULL`. For details, see
#'   the `Reproducible random number generation (RNG)` section below.
#'
#' @param scheduling A single integer, logical, or `Inf`. This argument
#'   controls the average number of futures ("chunks") per worker.
#'
#'   - If `0`, then a single future is used to process all elements of `.x`.
#'
#'   - If `1` or `TRUE`, then one future per worker is used.
#'
#'   - If `2`, then each worker will process two futures (provided there
#'     are enough elements in `.x`).
#'
#'   - If `Inf` or `FALSE`, then one future per element of `.x` is used.
#'
#'   This argument is only used if `chunk_size` is `NULL`.
#'
#' @param chunk_size A single integer, `Inf`, or `NULL`. This argument
#'   controls the average number of elements per future (`"chunk"`). If `Inf`,
#'   then all elements are processed in a single future. If `NULL`, then
#'   `scheduling` is used instead to determine how `.x` is chunked.
#'
#' @param prefix A single character string, or `NULL`. If a character string,
#'   then each future is assigned a label as `{prefix}-{chunk-id}`. If `NULL`,
#'   no labels are used.
#'
#'
#' @section Global variables:
#'
#' `globals` controls how globals are identified, similar to the `globals`
#' argument of [future::future()]. Since all function calls use the same set of
#' globals, furrr gathers globals upfront (once), which is more efficient than
#' if it was done for each future independently.
#'
#' * If `TRUE` or `NULL`, then globals are automatically identified and
#'   gathered.
#'
#' * If a character vector of names is specified, then those globals are
#'   gathered.
#'
#' * If a named list, then those globals are used as is.
#'
#' * In all cases, `.f` and any `...` arguments are automatically passed as
#'   globals to each future created, as they are always needed.
#'
#' @section Reproducible random number generation (RNG):
#'
#' Unless `seed = FALSE`, furrr functions are guaranteed to generate
#' the exact same sequence of random numbers _given the same initial
#' seed / RNG state_ regardless of the type of futures and scheduling
#' ("chunking") strategy.
#'
#' Setting `seed = NULL` is equivalent to `seed = FALSE`, except that the
#' `future.rng.onMisuse` option is not consulted to potentially monitor the
#' future for faulty random number usage. See the `seed` argument of
#' [future::future()] for more details.
#'
#' RNG reproducibility is achieved by pre-generating the random seeds for all
#' iterations (over `.x`) by using L'Ecuyer-CMRG RNG streams. In each
#' iteration, these seeds are set before calling `.f(.x[[i]], ...)`.
#' _Note, for large `length(.x)` this may introduce a large overhead._
#'
#' A fixed `seed` may be given as an integer vector, either as a full
#' L'Ecuyer-CMRG RNG seed of length `7`, or as a seed of length `1` that
#' will be used to generate a full L'Ecuyer-CMRG seed.
#'
#' If `seed = TRUE`, then `.Random.seed` is returned if it holds a
#' L'Ecuyer-CMRG RNG seed, otherwise one is created randomly.
#'
#' If `seed = NA`, a L'Ecuyer-CMRG RNG seed is randomly created.
#'
#' If none of the function calls `.f(.x[[i]], ...)` use random number
#' generation, then `seed = FALSE` may be used.
#'
#' In addition to the above, it is possible to specify a pre-generated
#' sequence of RNG seeds as a list such that `length(seed) == length(.x)` and
#' where each element is an integer seed that can be assigned to `.Random.seed`.
#' Use this alternative with caution. _Note that `as.list(seq_along(.x))` is
#' not a valid set of such `.Random.seed` values._
#'
#' In all cases but `seed = FALSE`, after a furrr function returns, the RNG
#' state of the calling R process is guaranteed to be "forwarded one step" from
#' the RNG state before the call. This is true regardless of the future
#' strategy / scheduling used. This is done in order to guarantee that an R
#' script calling `future_map()` multiple times should be numerically
#' reproducible given the same initial seed.
#'
#' Note that you cannot expect identical results between `map()` and
#' `future_map()` when using a `.f` that calls functions that generate random
#' numbers, even when calling `set.seed()` ahead of time. For one thing, the
#' default random number generation algorithm used by R during sequential
#' processing is Mersenne-Twister, different from the L'Ecuyer-CMRG seeds used
#' by furrr. But even aligning the [RNGkind()] would not be enough. `map()`
#' itself would have to change to use the same parallel compatible RNG strategy
#' as `future_map()` (pre-generating the seeds, and setting them before each
#' `.f` invocation). At the end of the day, you have to accept that the
#' following will produce different sequences of random numbers, but both are
#' statistically sound:
#'
#' ```r
#' set.seed(42)
#' purrr::map(1:10, ~ rnorm(1))
#'
#' set.seed(42)
#' furrr::future_map(1:10, ~ rnorm(1), .options = furrr_options(seed = TRUE))
#' ```
#'
#' But importantly, the `furrr::future_map()` example will always produce the
#' same sequence of random numbers, regardless of the `plan()` you choose:
#'
#' ```r
#' plan(sequential)
#' set.seed(42)
#' furrr::future_map(1:10, ~ rnorm(1), .options = furrr_options(seed = TRUE))
#'
#' plan(multisession, workers = 2)
#' set.seed(42)
#' furrr::future_map(1:10, ~ rnorm(1), .options = furrr_options(seed = TRUE))
#'
#' plan(cluster, workers = workers)
#' set.seed(42)
#' furrr::future_map(1:10, ~ rnorm(1), .options = furrr_options(seed = TRUE))
#' ```
#'
#' @export
#' @examples
#' furrr_options()
furrr_options <- function(
  ...,
  stdout = TRUE,
  conditions = "condition",
  globals = TRUE,
  packages = NULL,
  seed = FALSE,
  scheduling = 1.0,
  chunk_size = NULL,
  prefix = NULL
) {
  check_dots_empty()

  stdout <- validate_stdout(stdout)
  conditions <- validate_conditions(conditions)
  globals <- validate_globals(globals)
  packages <- validate_packages(packages)
  seed <- validate_seed(seed)
  scheduling <- validate_scheduling(scheduling)
  chunk_size <- validate_chunk_size(chunk_size)
  prefix <- validate_prefix(prefix)

  out <- list(
    stdout = stdout,
    conditions = conditions,
    globals = globals,
    packages = packages,
    seed = seed,
    scheduling = scheduling,
    chunk_size = chunk_size,
    prefix = prefix
  )

  structure(out, class = "furrr_options")
}

#' @export
print.furrr_options <- function(x, ...) {
  cat_line("<furrr_options>")
}

# ------------------------------------------------------------------------------

assert_furrr_options <- function(x) {
  if (!is_furrr_options(x)) {
    abort("`.options` must be created from `furrr_options()`.")
  }

  invisible(x)
}

is_furrr_options <- function(x) {
  inherits(x, "furrr_options")
}

# ------------------------------------------------------------------------------

validate_stdout <- function(x) {
  vctrs::vec_assert(x, size = 1L, arg = "stdout")

  # Allowed to be `NA`, as this means output is not intercepted.
  # We test for this but it is explicitly not recommended by future so we don't
  # document it.
  if (!is.logical(x)) {
    abort("`stdout` must be `TRUE`, `FALSE`, or `NA`.")
  }

  # Always drop stdout from the future objects.
  # Only the values are ever returned, so it is useless and potentially
  # expensive to transfer this back from the workers (#216).
  attr(x, "drop") <- TRUE

  x
}

validate_conditions <- function(x) {
  if (is.null(x)) {
    return(NULL)
  }

  if (!is.character(x)) {
    abort("`conditions` must be a character vector.")
  }

  # Always drop conditions from the future objects.
  # Only the values are ever returned, so it is useless and potentially
  # expensive to transfer these back from the workers (#216).
  attr(x, "drop") <- TRUE

  x
}

validate_globals <- function(x) {
  if (is.null(x)) {
    return(TRUE)
  }

  if (is.logical(x)) {
    if (!is_bool(x)) {
      abort("A logical `globals` must be length 1 and can't be `NA`.")
    }

    return(x)
  }

  if (is.character(x)) {
    return(x)
  }

  if (is.list(x)) {
    if (!is_named(x)) {
      abort("`globals` cannot be an unnamed or partially named list.")
    }

    return(x)
  }

  abort(
    "`globals` must be `NULL`, a logical, a character vector, or a named list."
  )
}

validate_packages <- function(x) {
  if (is.null(x)) {
    return(x)
  }

  x <- vctrs::vec_cast(x, character(), x_arg = "packages")

  if (any(is.na(x))) {
    abort("`packages` can't be `NA`.")
  }

  x
}

validate_seed <- function(x) {
  if (is.null(x)) {
    return(x)
  }

  # Size is validated when we have `.x`
  if (is.list(x)) {
    x <- validate_seed_list(x)
    return(x)
  }

  if (is.logical(x)) {
    if (!is_bool(x)) {
      abort("A logical `seed` must be length 1 and must not be `NA`.")
    }

    return(x)
  }

  if (is.double(x)) {
    x <- vctrs::vec_cast(x, integer(), x_arg = "seed")
  }

  if (is.integer(x)) {
    if (length(x) != 1L && length(x) != 7L) {
      abort("An integer `seed` must be length 1 or 7.")
    }

    if (any(is.na(x))) {
      abort("An integer `seed` cannot have `NA` values.")
    }

    return(x)
  }

  abort("`seed` must be a logical, integer, or list.")
}

validate_seed_list <- function(x) {
  seeds_are_integers <- purrr::map_lgl(x, ~ typeof(.x) == "integer")
  if (!all(seeds_are_integers)) {
    abort("All elements of a list `seed` must be integers.")
  }

  unique_lengths <- unique(lengths(x))
  if (length(unique_lengths) != 1L) {
    abort("All elements of a list `seed` must have the same length.")
  }

  if (identical(unique_lengths, 1L)) {
    abort(
      paste0(
        "All pre-generated random seed elements of a list `seed` ",
        "must be valid `.Random.seed` seeds, which means they should be all ",
        "integers and consists of two or more elements, not just one."
      )
    )
  }

  # For efficiency, just check the first seed for validity
  seed <- x[[1]]
  if (!is_valid_random_seed(seed)) {
    abort(
      paste0(
        "All pre-generated random seed elements of a list `seed` ",
        "must be valid `.Random.seed` seeds, but the first does not seem to be."
      )
    )
  }

  x
}

validate_scheduling <- function(x) {
  vctrs::obj_check_vector(x, arg = "scheduling")
  vctrs::vec_check_size(x, size = 1L, arg = "scheduling")

  if (is.na(x)) {
    abort("`scheduling` can't be `NA`.")
  }

  if (is_bool(x)) {
    return(x)
  }

  if (x < 0L) {
    abort("`scheduling` must be greater than or equal to zero.")
  }

  if (is.infinite(x)) {
    return(x)
  }

  ordering <- attr(x, "ordering")
  x <- vctrs::vec_cast(x, integer(), x_arg = "scheduling")
  attr(x, "ordering") <- ordering

  x
}

validate_chunk_size <- function(x) {
  if (is.null(x)) {
    return(x)
  }

  vctrs::obj_check_vector(x, arg = "chunk_size")
  vctrs::vec_check_size(x, size = 1L, arg = "chunk_size")

  if (is.na(x)) {
    abort("`chunk_size` can't be `NA`.")
  }

  if (x <= 0L) {
    abort("`chunk_size` must be greater than zero.")
  }

  if (is.infinite(x)) {
    return(x)
  }

  ordering <- attr(x, "ordering")
  x <- vctrs::vec_cast(x, integer(), x_arg = "chunk_size")
  attr(x, "ordering") <- ordering

  x
}

validate_prefix <- function(x) {
  if (is.null(x)) {
    return(x)
  }

  if (!is_string(x)) {
    abort("`prefix` must be a character string, or `NULL`.")
  }

  x
}

#' Map over multiple inputs simultaneously via futures
#'
#' These functions work the same as [purrr::map2()] and its variants,
#' but allow you to map in parallel. Note that "parallel" as described in purrr
#' is just saying that you are working with multiple inputs, and parallel in
#' this case means that you can work on multiple inputs and process them all in
#' parallel as well.
#'
#' @inheritParams future_map
#'
#' @param .x,.y A pair of vectors, usually the same length. If not, a vector
#'   of length 1 will be recycled to the length of the other.
#'
#' @param .l A list of vectors. The length of `.l` determines the number of
#'   arguments that `.f` will be called with. Arguments will be supply by
#'   position if unnamed, and by name if named.
#'
#'   Vectors of length 1 will be recycled to any length; all other elements
#'   must be have the same length.
#'
#'   A data frame is an important special case of `.l`. It will cause `.f`
#'   to be called once for each row.
#'
#' @param .f A function, specified in one of the following ways:
#'
#'   * A named function.
#'   * An anonymous function, e.g. `\(x, y) x + y` or `function(x, y) x + y`.
#'   * A formula, e.g. `~ .x + .y`. Use `.x` to refer to the current
#'     element of `x` and `.y` to refer to the current element of `y`.
#'     No longer recommended.
#'
#' @return An atomic vector, list, or data frame, depending on the suffix.
#'   Atomic vectors and lists will be named if `.x` or the first element of `.l`
#'   is named.
#'
#'   If all input is length 0, the output will be length 0. If any input is
#'   length 1, it will be recycled to the length of the longest.
#'
#' @export
#' @examples
#' \donttest{plan(multisession, workers = 2)}
#'
#' x <- list(1, 10, 100)
#' y <- list(1, 2, 3)
#' z <- list(5, 50, 500)
#'
#' future_map2(x, y, ~ .x + .y)
#'
#' # Split into pieces, fit model to each piece, then predict
#' by_cyl <- split(mtcars, mtcars$cyl)
#' mods <- future_map(by_cyl, ~ lm(mpg ~ wt, data = .))
#' future_map2(mods, by_cyl, predict)
#'
#' future_pmap(list(x, y, z), sum)
#'
#' # Matching arguments by position
#' future_pmap(list(x, y, z), function(a, b ,c) a / (b + c))
#'
#' # Vectorizing a function over multiple arguments
#' df <- data.frame(
#'   x = c("apple", "banana", "cherry"),
#'   pattern = c("p", "n", "h"),
#'   replacement = c("x", "f", "q"),
#'   stringsAsFactors = FALSE
#' )
#'
#' future_pmap(df, gsub)
#' future_pmap_chr(df, gsub)
#'
#' \dontshow{
#' # Close open connections for R CMD Check
#' if (!inherits(plan(), "sequential")) plan(sequential)
#' }
future_map2 <- function(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map2_template(
    x = .x,
    y = .y,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = "map2",
    env_globals = .env_globals
  )
}

#' @rdname future_map2
#' @export
future_map2_chr <- function(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map2_template(
    x = .x,
    y = .y,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "character",
    purrr_fn_name = "map2_chr",
    env_globals = .env_globals
  )
}

#' @rdname future_map2
#' @export
future_map2_dbl <- function(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map2_template(
    x = .x,
    y = .y,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "double",
    purrr_fn_name = "map2_dbl",
    env_globals = .env_globals
  )
}

#' @rdname future_map2
#' @export
future_map2_int <- function(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map2_template(
    x = .x,
    y = .y,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "integer",
    purrr_fn_name = "map2_int",
    env_globals = .env_globals
  )
}

#' @rdname future_map2
#' @export
future_map2_lgl <- function(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_map2_template(
    x = .x,
    y = .y,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "logical",
    purrr_fn_name = "map2_lgl",
    env_globals = .env_globals
  )
}

#' @rdname future_map2
#' @export
future_map2_vec <- function(
  .x,
  .y,
  .f,
  ...,
  .ptype = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  out <- future_map2(
    .x = .x,
    .y = .y,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  simplify_impl(
    out,
    ptype = .ptype,
    error_arg = "<output>",
    error_call = current_env()
  )
}

#' @rdname future_map2
#' @export
future_map2_dfr <- function(
  .x,
  .y,
  .f,
  ...,
  .id = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  if (!rlang::is_installed("dplyr")) {
    rlang::abort("`future_map2_dfr()` requires dplyr")
  }

  res <- future_map2(
    .x = .x,
    .y = .y,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  dplyr::bind_rows(res, .id = .id)
}

#' @rdname future_map2
#' @export
future_map2_dfc <- function(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  if (!rlang::is_installed("dplyr")) {
    rlang::abort("`future_map2_dfc()` requires dplyr")
  }

  res <- future_map2(
    .x = .x,
    .y = .y,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  dplyr::bind_cols(res)
}

# ------------------------------------------------------------------------------
# map2()

furrr_test_that("future_map2() matches map2() for simple cases", {
  expect_identical(
    future_map2(1:3, 4:6, ~ .x + .y),
    map2(1:3, 4:6, ~ .x + .y)
  )
})

furrr_test_that("names of `.x` are retained", {
  x <- c(a = 1, b = 2)
  y <- c(c = 1, d = 2)
  expect_named(future_map2(x, y, ~1), c("a", "b"))
})

furrr_test_that("named empty input makes named empty output", {
  x <- set_names(list(), character())
  expect_named(future_map2(x, x, ~.x), character())
})

# ------------------------------------------------------------------------------
# atomic variants

furrr_test_that("future_map2_dbl() works", {
  x <- c(1, 2, 3)
  y <- c(4, 5, 6)

  expect_identical(
    future_map2_dbl(x, y, ~ .x + .y),
    map2_dbl(x, y, ~ .x + .y)
  )
})

furrr_test_that("future_map2_int() works", {
  x <- c(1L, 2L, 3L)
  y <- c(4L, 5L, 6L)

  expect_identical(
    future_map2_int(x, y, ~ .x + .y),
    map2_int(x, y, ~ .x + .y)
  )
})

furrr_test_that("future_map2_lgl() works", {
  x <- c(TRUE, FALSE, TRUE)
  y <- c(FALSE, TRUE, TRUE)

  expect_identical(
    future_map2_lgl(x, y, ~ .x || .y),
    map2_lgl(x, y, ~ .x || .y)
  )
})

furrr_test_that("future_map2_chr() works", {
  x <- c("a", "b", "c")
  y <- c("d", "e", "f")

  expect_identical(
    future_map2_chr(x, y, ~.y),
    map2_chr(x, y, ~.y)
  )
})

furrr_test_that("future_map2_vec() works", {
  x <- as.Date(c("2020-01-01", "2020-01-02", "2020-01-03"))
  y <- 1:3

  expect_identical(
    future_map2_vec(x, y, ~.x),
    map2_vec(x, y, ~.x)
  )
  expect_identical(
    future_map2_vec(x, y, ~1, .ptype = integer()),
    map2_vec(x, y, ~1, .ptype = integer())
  )

  expect_identical(
    future_map2_vec(integer(), integer(), identity),
    map2_vec(integer(), integer(), identity)
  )
  expect_identical(
    future_map2_vec(
      set_names(integer(), character()),
      integer(),
      identity
    ),
    map2_vec(
      set_names(integer(), character()),
      integer(),
      identity
    )
  )

  # Vector error
  expect_snapshot(error = TRUE, {
    future_map2_vec(1:2, 1:2, ~NULL)
  })
  # Size error
  expect_snapshot(error = TRUE, {
    future_map2_vec(1:2, 1:2, ~ 1:2)
  })
  # Type error
  expect_snapshot(error = TRUE, {
    future_map2_vec(1:2, 1:2, ~ if (.x == 1L) 1 else "x")
  })
})

furrr_test_that("names of `.x` are retained", {
  x <- c(a = 1, b = 2)
  y <- c(c = 1, d = 2)
  expect_named(future_map2_dbl(x, y, ~1), c("a", "b"))

  x <- c(a = as.Date("2020-01-01"), b = as.Date("2020-01-02"))
  y <- c(c = 1, d = 2)
  expect_named(future_map2_vec(x, y, ~1), c("a", "b"))
})

# ------------------------------------------------------------------------------
# data frame variants

furrr_test_that("future_map2_dfr() works", {
  x <- c("a", "b", "c")
  y <- c("d", "e", "f")

  expect_identical(
    future_map2_dfr(x, y, ~ data.frame(x = .x, y = .y)),
    map2_dfr(x, y, ~ data.frame(x = .x, y = .y))
  )
})

furrr_test_that("future_map2_dfc() works", {
  x <- c("a", "b", "c")
  y <- c("d", "e", "f")

  expect_identical(
    future_map2_dfc(x, y, ~ as.data.frame(set_names(list(.x), .y))),
    map2_dfc(x, y, ~ as.data.frame(set_names(list(.x), .y)))
  )
})

# ------------------------------------------------------------------------------
# size

furrr_test_that("future_map2() works with size zero input", {
  expect_identical(future_map2(list(), list(), identity), list())
})

furrr_test_that("atomic variants work with size zero input", {
  expect_identical(future_map2_chr(list(), list(), identity), character())
  expect_identical(future_map2_dbl(list(), list(), identity), double())
  expect_identical(future_map2_int(list(), list(), identity), integer())
  expect_identical(future_map2_lgl(list(), list(), identity), logical())
})

furrr_test_that("generic variant works with size zero input", {
  expect_identical(
    future_map2_vec(list(), list(), identity),
    NULL
  )
  expect_identical(
    future_map2_vec(list(), list(), identity, .ptype = integer()),
    integer()
  )
})

furrr_test_that("size one recycling works", {
  expect_identical(
    future_map2(1, 1:2, ~ c(.x, .y)),
    list(c(1, 1), c(1, 2))
  )

  expect_identical(
    future_map2(1:2, 1, ~ c(.x, .y)),
    list(c(1, 1), c(2, 1))
  )

  expect_identical(
    future_map2(integer(), 1, ~ c(.x, .y)),
    list()
  )

  expect_identical(
    future_map2(1, integer(), ~ c(.x, .y)),
    list()
  )
})

# TODO: Reenable this test after future issue is fixed
# https://github.com/futureverse/future/issues/820
# https://github.com/futureverse/furrr/issues/307
# furrr_test_that("generally can't recycle to size zero", {
#   expect_error(
#     future_map2(1:2, integer(), ~ c(.x, .y)),
#     "Can't recycle"
#   )
#
#   expect_error(
#     future_map2(integer(), 1:2, ~ c(.x, .y)),
#     "Can't recycle"
#   )
# })

# ------------------------------------------------------------------------------
# Miscellaneous

furrr_test_that("globals in `.x` and `.y` are found (#16)", {
  fn1 <- function(x) sum(x, na.rm = TRUE)
  fn2 <- function(x) sum(x, na.rm = FALSE)

  x <- list(c(1, 2, NA), c(2, 3, 4))

  fns1 <- map(x, ~ purrr::partial(fn1, x = .x))
  fns2 <- map(x, ~ purrr::partial(fn2, x = .x))

  expect_identical(
    future_map2(fns1, fns2, ~ c(.x(), .y())),
    list(c(3, NA), c(9, 9))
  )
})

furrr_test_that("chunk balancing is correct after a recycle (#30)", {
  expect_identical(
    future_map2(1, 1:4, ~ c(.x, .y)),
    list(c(1, 1), c(1, 2), c(1, 3), c(1, 4))
  )
})

# nocov start

probe <- function(.x, .p, ...) {
  if (is_logical(.p)) {
    stopifnot(length(.p) == length(.x))
    .p
  } else {
    purrr::map_lgl(.x, .p, ...)
  }
}

inv_which <- function(x, sel) {
  if (is.character(sel)) {
    names <- names(x)
    if (is.null(names)) {
      stop("character indexing requires a named object", call. = FALSE)
    }
    names %in% sel
  } else if (is.numeric(sel)) {
    if (any(sel < 0)) {
      !seq_along(x) %in% abs(sel)
    } else {
      seq_along(x) %in% sel
    }
  } else {
    stop("unrecognised index type", call. = FALSE)
  }
}

vec_index <- function(x) {
  names(x) %||% seq_along(x)
}

check_tidyselect <- function() {
  if (!is_installed("tidyselect")) {
    abort("Using tidyselect in `future_map_at()` requires tidyselect")
  }
}

at_selection <- function(nm, .at) {
  if (is_quosures(.at)) {
    check_tidyselect()
    .at <- tidyselect::vars_select(.vars = nm, !!!.at)
  }
  .at
}

# `purrr:::simplify_impl()`, but simplified (ha!) by removing `strict`,
# since we are always `strict`
simplify_impl <- function(x, ptype, error_arg, error_call) {
  vctrs::obj_check_list(x, arg = error_arg, call = error_call)
  vctrs::list_check_all_vectors(x, arg = error_arg, call = error_call)
  vctrs::list_check_all_size(x, 1L, arg = error_arg, call = error_call)

  names <- vctrs::vec_names(x)
  x <- vctrs::vec_set_names(x, NULL)

  out <- vctrs::vec_c(
    !!!x,
    .ptype = ptype,
    .error_arg = error_arg,
    .error_call = error_call
  )

  if (!is.null(out)) {
    out <- vctrs::vec_set_names(out, names)
  }

  out
}

# nocov end

# nocov start

poll_progress <- function(futures, file, n_x) {
  symbol <- get_progress_symbol()

  prefix <- "Progress: "
  suffix <- " 100%"

  width_prefix <- nchar(prefix)
  width_suffix <- nchar(suffix)
  width_carriage <- 1L

  stderr <- stderr()

  while (any_running(futures)) {
    con <- file(file, open = "r")
    n_ticks <- get_n_ticks(con)
    close(con)

    # Console width might change while we poll
    width_max <- console_width()
    width_usable <- width_max - width_prefix - width_suffix - width_carriage

    # Supposedly someone has seen the multiplication integer overflow for
    # extremely long inputs, so we `as.double()` one of the operands, but the
    # result after division should still always fit in an integer (#288)
    width_rule <- as.integer(floor(width_usable * as.double(n_ticks) / n_x))
    width_space <- width_usable - width_rule

    space <- paste0(rep(" ", times = width_space), collapse = "")
    rule <- paste0(rep(symbol, times = width_rule), collapse = "")

    out <- paste0(prefix, rule, space, suffix)

    cat("\r", out, file = stderr)
    utils::flush.console()
  }
}

get_n_ticks <- function(con) {
  line <- readLines(con, n = 1L, warn = FALSE)

  if (length(line) == 0L) {
    line <- ""
  }

  nchar(line)
}

any_running <- function(futures) {
  !all(future::resolved(futures))
}

console_width <- function() {
  width <- Sys.getenv("RSTUDIO_CONSOLE_WIDTH", getOption("width", 80))
  as.integer(width)
}

# ------------------------------------------------------------------------------

# Adapted from cli's onload properties
# to dynamically switch depending on utf8 availability
get_progress_symbol <- function() {
  if (is_utf8_output()) {
    "\u2500"
  } else {
    "-"
  }
}

is_utf8_output <- function() {
  l10n_info()$`UTF-8` && !is_latex_output()
}

is_latex_output <- function() {
  if (!("knitr" %in% loadedNamespaces())) {
    return(FALSE)
  }

  get("is_latex_output", asNamespace("knitr"))()
}

# ------------------------------------------------------------------------------

assert_progress <- function(progress) {
  if (!is_bool(progress)) {
    abort("`.progress` must be a single logical value.")
  }

  invisible(progress)
}

# ------------------------------------------------------------------------------

# - Sequential blocks in the `future()` call, so no progress is ever shown
# - Cluster is generally used for multi-computer setups, and would end up
#   writing into files on the remote workers, which would never be shown.
progress_enabled_plans <- c(
  "multicore",
  "multisession",
  "multiprocess"
)

reconcile_progress_with_strategy <- function(progress) {
  if (is_false(progress)) {
    return(progress)
  }

  plan <- future::plan()
  progress_enabled_plan <- inherits_any(plan, progress_enabled_plans)

  if (!progress_enabled_plan) {
    progress <- FALSE
  }

  progress
}

# nocov end

furrr_recycle_common <- function(x, n) {
  n_x <- length(x)
  lengths <- lengths(x)
  idx <- rep_len(1L, n)

  for (i in seq_len(n_x)) {
    elt <- x[[i]]

    # Don't recycle `NULL` elements, they can be indexed fine by `[` and `[[`
    # and are considered like missing arguments
    if (is.null(elt)) {
      next
    }

    length_elt <- lengths[[i]]

    if (length_elt == n) {
      next
    }

    if (length_elt == 1L) {
      x[[i]] <- elt[idx]
      next
    }

    abort(paste0("Internal error: Incompatible lengths at location ", i, "."))
  }

  x
}

# Can't use `vec_size_common()` because we extract elements with `[[` and
# respect length invariants, not size invariants
furrr_length_common <- function(x) {
  # Don't consider `NULL` elements in common size
  x <- compact_null(x)

  # Handle empty pmap input
  if (length(x) == 0L) {
    return(0L)
  }

  lengths <- lengths(x)
  indices <- seq_along(lengths)

  purrr::reduce2(lengths, indices, furrr_length2, .init = 1L)
}

furrr_length2 <- function(x, y, i) {
  if (x == 1L) {
    y
  } else if (y == 1L) {
    x
  } else if (x == y) {
    x
  } else {
    msg <- paste0(
      "Can't recycle length ",
      x,
      " and length ",
      y,
      " at location ",
      i,
      "."
    )
    abort(msg)
  }
}

compact_null <- function(x) {
  null <- purrr::map_lgl(x, is.null)

  if (any(null)) {
    x[!null]
  } else {
    x
  }
}

cat_line <- function(...) {
  cat(paste0(..., "\n", collapse = ""))
}

import_future <- function(name, default = NULL) {
  import_from(name, default = default, package = "future")
}

import_from <- function(name, default = NULL, package) {
  ns <- getNamespace(package)
  if (exists(name, mode = "function", envir = ns, inherits = FALSE)) {
    get(name, mode = "function", envir = ns, inherits = FALSE)
  } else if (!is.null(default)) {
    default
  } else {
    stop(sprintf("No such '%s' function: %s()", package, name))
  }
}

get_total_size <- function(x) {
  attr(x, "total_size", exact = TRUE)
}
set_total_size <- function(x, total_size) {
  attr(x, "total_size") <- total_size
  x
}

# Adapted from:
# https://github.com/HenrikBengtsson/future.apply/blob/develop/R/rng.R

make_seeds <- function(seed, n) {
  if (is.list(seed)) {
    seed <- validate_supplied_seeds(seed, n)
    return(seed)
  }

  if (is_true(seed)) {
    seed <- as_lecyer_cmrg_seed_from_base_r_seed()
  } else if (is.integer(seed)) {
    seed <- as_lecyer_cmrg_seed_from_integer(seed)
  } else {
    abort("Internal error: Unknown type of `seed` encountered.")
  }

  oseed <- next_random_seed()
  on.exit(set_random_seed(oseed), add = TRUE)

  out <- vector("list", length = n)

  for (i in seq_len(n)) {
    out[[i]] <- nextRNGSubStream(seed)
    seed <- nextRNGStream(seed)
  }

  out
}

# ------------------------------------------------------------------------------

validate_supplied_seeds <- function(seeds, n) {
  # Finally check length of `seed` against `n`
  if (length(seeds) != n) {
    abort(
      paste0(
        "If `furrr_options(seed = )` is a list, it must have length equal ",
        "to the common length of the inputs, ",
        n,
        ", ",
        "not length ",
        length(seeds),
        "."
      )
    )
  }

  # All other checks on a list `seed` were done in `furrr_options()`
  seeds
}

# ------------------------------------------------------------------------------

is_valid_random_seed <- function(seed) {
  oseed <- get_random_seed()
  on.exit(set_random_seed(oseed), add = TRUE)

  env <- globalenv()
  env$.Random.seed <- seed

  res <- tryCatch(
    simpleWarning = function(cnd) cnd,
    sample.int(n = 1L, size = 1L, replace = FALSE)
  )

  ok <- !inherits(res, "simpleWarning")

  ok
}

next_random_seed <- function(seed = get_random_seed()) {
  sample.int(n = 1L, size = 1L, replace = FALSE)
  seed_next <- get_random_seed()

  stopifnot(identical(seed_next, seed))

  invisible(seed_next)
}

get_random_seed <- function() {
  env <- globalenv()
  env$.Random.seed
}

set_random_seed <- function(seed) {
  env <- globalenv()

  if (is.null(seed)) {
    rm(list = ".Random.seed", envir = env, inherits = FALSE)
  } else {
    env$.Random.seed <- seed
  }
}

# ------------------------------------------------------------------------------

as_lecyer_cmrg_seed_from_base_r_seed <- function() {
  oseed <- get_random_seed()

  # Already a L'Ecuyer-CMRG seed?  Then use that as is.
  if (is_lecyer_cmrg_seed(oseed)) {
    return(oseed)
  }

  on.exit(set_random_seed(oseed), add = TRUE)

  RNGkind("L'Ecuyer-CMRG")

  return(get_random_seed())
}

as_lecyer_cmrg_seed_from_integer <- function(seed) {
  # Already a L'Ecuyer-CMRG seed?
  if (is_lecyer_cmrg_seed(seed)) {
    return(seed)
  }

  # Generate a new L'Ecuyer-CMRG seed
  if (length(seed) == 1L) {
    oseed <- get_random_seed()
    on.exit(set_random_seed(oseed), add = TRUE)

    RNGkind("L'Ecuyer-CMRG")
    set.seed(seed)

    return(get_random_seed())
  }

  abort(
    paste0(
      "Integer `seed` must be L'Ecuyer-CMRG RNG seed as returned by ",
      "`parallel::nextRNGStream()` or a single integer."
    )
  )
}

is_lecyer_cmrg_seed <- function(seed) {
  is.numeric(seed) &&
    length(seed) == 7L &&
    all(is.finite(seed)) &&
    (seed[[1]] %% 10000L == 407L)
}

test_that("checks list of seeds for length", {
  old_rngkind <- RNGkind("L'Ecuyer-CMRG")
  on.exit(RNGkind(old_rngkind[[1]]), add = TRUE)

  lecuyer_seed <- get_random_seed()

  x <- list(lecuyer_seed, lecuyer_seed, lecuyer_seed)
  y <- list(lecuyer_seed)

  expect_identical(make_seeds(x, 3), x)
  expect_error(make_seeds(y, 3), "3, not length 1")
})

test_that("can make `seeds` from `TRUE` `seed`", {
  x <- make_seeds(TRUE, 5)
  expect_identical(length(x), 5L)
  expect_true(all(purrr::map_int(x, length) == 7L))
})

test_that("can make reproducible `seeds` from integer `seed`", {
  x <- make_seeds(1L, 5)
  y <- make_seeds(1L, 5)
  expect_identical(length(x), 5L)
  expect_true(all(purrr::map_int(x, length) == 7L))
  expect_identical(x, y)
})

make_chunks <- function(n_x, n_workers, scheduling = 1L, chunk_size = NULL) {
  if (is.null(chunk_size)) {
    n_chunks <- compute_n_chunks_from_scheduling(scheduling, n_x, n_workers)
  } else {
    n_chunks <- compute_n_chunks_from_chunk_size(chunk_size, n_x)
  }

  parallel::splitIndices(n_x, n_chunks)
}

compute_n_chunks_from_scheduling <- function(scheduling, n_x, n_workers) {
  if (is.logical(scheduling)) {
    compute_n_chunks_from_scheduling_logical(scheduling, n_x, n_workers)
  } else if (is.integer(scheduling)) {
    compute_n_chunks_from_scheduling_integer(scheduling, n_x, n_workers)
  } else if (identical(scheduling, Inf)) {
    n_x
  }
}

compute_n_chunks_from_scheduling_logical <- function(
  scheduling,
  n_x,
  n_workers
) {
  # One future per element of `x`
  if (is_false(scheduling)) {
    return(n_x)
  }

  # One future per worker (if there are enough elements in `x`)
  if (n_workers > n_x) {
    n_x
  } else {
    n_workers
  }
}

# Treat:
# N chunks / 1 workers = scheduling
# So:
# N chunks = (N chunks / 1 workers) * N workers
compute_n_chunks_from_scheduling_integer <- function(
  scheduling,
  n_x,
  n_workers
) {
  if (n_workers > n_x) {
    n_workers <- n_x
  }

  n_chunks <- scheduling * n_workers

  # Catch `scheduling == 0L`
  if (n_chunks < 1L) {
    n_chunks <- 1L
  }

  if (n_chunks > n_x) {
    n_chunks <- n_x
  }

  n_chunks
}

# From `future.apply:::makeChunks()`, with description of:
# "Same definition as parallel:::staticNChunks() in R (>= 3.5.0)"
compute_n_chunks_from_chunk_size <- function(chunk_size, n_x) {
  max(1L, ceiling(n_x / chunk_size))
}

# ------------------------------------------------------------------------------

# Custom ordering can be used to process the elements of `x` in. It is
# extracted from `scheduling` or `chunk_size` as the `"ordering"` attribute.
make_order <- function(n_x, scheduling, chunk_size) {
  if (is.null(chunk_size)) {
    ordering <- attr(scheduling, "ordering", exact = TRUE)
    compute_order(ordering, n_x, "scheduling")
  } else {
    ordering <- attr(chunk_size, "ordering", exact = TRUE)
    compute_order(ordering, n_x, "chunk_size")
  }
}

compute_order <- function(ordering, n_x, arg) {
  if (is.null(ordering)) {
    return(NULL)
  }

  if (is.character(ordering) && identical(ordering, "random")) {
    order <- sample_n_with_reset(n_x)
  } else if (is.integer(ordering)) {
    order <- ordering
  } else if (is.double(ordering)) {
    order <- vctrs::vec_cast(ordering, integer(), x_arg = "ordering")
  } else if (is.function(ordering)) {
    order <- ordering(n_x)
  } else {
    abort(
      paste0(
        "Unknown type of `",
        arg,
        "` attribute, `ordering`. The attribute must ",
        "be a character string, an integer, or a function."
      )
    )
  }

  if (!is.integer(order)) {
    abort("Order computed from `ordering` attribute must be an integer.")
  }

  if (length(order) != n_x) {
    abort("Order computed from `ordering` must have length equal to `.x`.")
  }

  order
}

# Reset on exit to not modify randomness state
sample_n_with_reset <- function(n) {
  oseed <- .GlobalEnv$.Random.seed

  on.exit({
    if (is.null(oseed)) {
      rm(list = ".Random.seed", envir = .GlobalEnv, inherits = FALSE)
    } else {
      .GlobalEnv$.Random.seed <- oseed
    }
  })

  sample.int(n = n, size = n, replace = FALSE)
}

test_that("can make chunks from logical `scheduling`", {
  expect_identical(
    make_chunks(n_x = 6L, n_workers = 2L, scheduling = TRUE, chunk_size = NULL),
    list(1:3, 4:6)
  )

  expect_identical(
    make_chunks(n_x = 6L, n_workers = 2L, scheduling = TRUE, chunk_size = NULL),
    make_chunks(n_x = 6L, n_workers = 2L, scheduling = 1L, chunk_size = NULL)
  )

  expect_identical(
    make_chunks(n_x = 5L, n_workers = 3L, scheduling = TRUE, chunk_size = NULL),
    list(1:2, 3L, 4:5)
  )

  expect_identical(
    make_chunks(n_x = 5L, n_workers = 3L, scheduling = TRUE, chunk_size = NULL),
    make_chunks(n_x = 5L, n_workers = 3L, scheduling = 1L, chunk_size = NULL)
  )

  expect_identical(
    make_chunks(
      n_x = 6L,
      n_workers = 2L,
      scheduling = FALSE,
      chunk_size = NULL
    ),
    as.list(1:6)
  )

  expect_identical(
    make_chunks(
      n_x = 6L,
      n_workers = 2L,
      scheduling = FALSE,
      chunk_size = NULL
    ),
    make_chunks(n_x = 6L, n_workers = 2L, scheduling = Inf, chunk_size = NULL)
  )
})

test_that("can make chunks from integer `scheduling`", {
  expect_identical(
    make_chunks(n_x = 6L, n_workers = 2L, scheduling = 0L, chunk_size = NULL),
    list(1:6)
  )

  expect_length(
    make_chunks(n_x = 7L, n_workers = 3L, scheduling = 2L, chunk_size = NULL),
    6L
  )

  expect_identical(
    make_chunks(n_x = 3L, n_workers = 3L, scheduling = 2L, chunk_size = NULL),
    list(1L, 2L, 3L)
  )

  expect_identical(
    make_chunks(n_x = 20L, n_workers = 2L, scheduling = 2L, chunk_size = NULL),
    list(1:5, 6:10, 11:15, 16:20)
  )
})

test_that("can make chunks from Inf `scheduling`", {
  expect_identical(
    make_chunks(n_x = 6L, n_workers = 2L, scheduling = Inf, chunk_size = NULL),
    as.list(1:6)
  )
})

test_that("can make chunks from integer `chunk_size`", {
  expect_identical(
    make_chunks(n_x = 6L, n_workers = 2L, scheduling = 1L, chunk_size = 1L),
    as.list(1:6)
  )

  expect_identical(
    make_chunks(n_x = 5L, n_workers = 2L, scheduling = 1L, chunk_size = 2L),
    list(1:2, 3L, 4:5)
  )
})

test_that("can make chunks from Inf `chunk_size`", {
  expect_identical(
    make_chunks(n_x = 6L, n_workers = 2L, scheduling = 1L, chunk_size = Inf),
    list(1:6)
  )
})

test_that("`chunk_size` takes presedence over `scheduling`", {
  expect_identical(
    make_chunks(n_x = 6L, n_workers = 2L, scheduling = 2L, chunk_size = Inf),
    list(1:6)
  )
})

test_that("can use random `ordering` attribute on `scheduling`", {
  x <- structure(2L, ordering = "random")

  set.seed(123)

  o1 <- make_order(n_x = 6L, scheduling = x, chunk_size = NULL)

  expect_length(o1, 6L)
  expect_true(all(1:6 %in% o1))

  # making the order randomly resets the seed
  o2 <- make_order(n_x = 6L, scheduling = x, chunk_size = NULL)
  expect_identical(o1, o2)
})

test_that("can use integer `ordering` attribute", {
  x <- structure(2L, ordering = c(3L, 2L, 1L))

  o <- make_order(n_x = 3L, scheduling = x, chunk_size = NULL)

  expect_identical(o, attr(x, "ordering", exact = TRUE))
})

test_that("can use function `ordering` attribute", {
  ordering_fn <- function(n) {
    rev(seq_len(n))
  }

  x <- structure(2L, ordering = ordering_fn)

  o <- make_order(n_x = 3L, scheduling = x, chunk_size = NULL)

  expect_identical(o, 3:1)
})

test_that("can use random `ordering` attribute on `chunk_size`", {
  x <- structure(2L, ordering = "random")

  set.seed(123)

  o1 <- make_order(n_x = 6L, scheduling = 1L, chunk_size = x)

  expect_length(o1, 6L)
  expect_true(all(1:6 %in% o1))

  # making the order randomly resets the seed
  o2 <- make_order(n_x = 6L, scheduling = x, chunk_size = NULL)
  expect_identical(o1, o2)
})

furrr_test_that("ordering is undone at the end", {
  ordering_fn <- function(n) {
    rev(seq_len(n))
  }

  x <- structure(2L, ordering = ordering_fn)

  opts <- furrr_options(scheduling = x)

  expect_identical(future_map(1:10, ~.x, .options = opts), as.list(1:10))
})

test_that("validates ordering", {
  x <- structure(2L, ordering = "rand")
  expect_error(make_order(n_x = 6L, scheduling = 1L, chunk_size = x), "Unknown")

  x <- structure(2L, ordering = 1L)
  expect_error(
    make_order(n_x = 6L, scheduling = 1L, chunk_size = x),
    "length equal"
  )

  x <- structure(2L, ordering = 1.5)
  expect_error(make_order(n_x = 6L, scheduling = 1L, chunk_size = x))
})

furrr_map_template <- function(
  x,
  fn,
  dots,
  options,
  progress,
  type,
  purrr_fn_name,
  env_globals
) {
  n <- length(x)
  names <- names(x)

  assert_furrr_options(options)
  assert_progress(progress)

  progress <- reconcile_progress_with_strategy(progress)

  expr_seed_setup <- make_expr_seed_setup(options$seed)
  expr_seed_update <- make_expr_seed_update(options$seed)

  expr_progress_setup <- make_expr_progress_setup(progress)
  expr_progress_update <- make_expr_progress_update(progress)

  walk <- identical(purrr_fn_name, "walk")

  if (walk) {
    purrr_fn_name <- "map"
  }

  expr_out <- make_expr_out(walk)

  expr <- expr({
    ...furrr_chunk_x <- ...furrr_chunk_args

    !!expr_seed_setup
    !!expr_progress_setup

    ...furrr_fn_wrapper <- function(...) {
      !!expr_seed_update
      !!expr_progress_update

      ...furrr_out <- ...furrr_fn(...)

      !!expr_out
    }

    ...furrr_purrr_fn <- get(
      !!purrr_fn_name,
      envir = asNamespace("purrr"),
      mode = "function",
      inherits = FALSE
    )

    args <- list(
      .x = ...furrr_chunk_x,
      .f = ...furrr_fn_wrapper
    )

    args <- c(args, ...furrr_dots)

    do.call(...furrr_purrr_fn, args)
  })

  furrr_template(
    args = x,
    fn = fn,
    dots = dots,
    n = n,
    options = options,
    progress = progress,
    type = type,
    names = names,
    env_globals = env_globals,
    expr = expr,
    extract = furrr_map_extract
  )
}

furrr_map_extract <- function(x, i) {
  x[i]
}

# ------------------------------------------------------------------------------

furrr_map2_template <- function(
  x,
  y,
  fn,
  dots,
  options,
  progress,
  type,
  purrr_fn_name,
  env_globals
) {
  args <- list(x, y)

  n <- furrr_length_common(args)
  args <- furrr_recycle_common(args, n)

  # Get names after possible recycling
  names <- names(args[[1]])

  assert_furrr_options(options)
  assert_progress(progress)

  progress <- reconcile_progress_with_strategy(progress)

  expr_seed_setup <- make_expr_seed_setup(options$seed)
  expr_seed_update <- make_expr_seed_update(options$seed)

  expr_progress_setup <- make_expr_progress_setup(progress)
  expr_progress_update <- make_expr_progress_update(progress)

  walk <- identical(purrr_fn_name, "walk2")

  if (walk) {
    purrr_fn_name <- "map2"
  }

  expr_out <- make_expr_out(walk)

  expr <- expr({
    ...furrr_chunk_x <- ...furrr_chunk_args[[1]]
    ...furrr_chunk_y <- ...furrr_chunk_args[[2]]

    !!expr_seed_setup
    !!expr_progress_setup

    ...furrr_fn_wrapper <- function(...) {
      !!expr_seed_update
      !!expr_progress_update

      ...furrr_out <- ...furrr_fn(...)

      !!expr_out
    }

    ...furrr_purrr_fn <- get(
      !!purrr_fn_name,
      envir = asNamespace("purrr"),
      mode = "function",
      inherits = FALSE
    )

    args <- list(
      .x = ...furrr_chunk_x,
      .y = ...furrr_chunk_y,
      .f = ...furrr_fn_wrapper
    )

    args <- c(args, ...furrr_dots)

    do.call(...furrr_purrr_fn, args)
  })

  furrr_template(
    args = args,
    fn = fn,
    dots = dots,
    n = n,
    options = options,
    progress = progress,
    type = type,
    names = names,
    env_globals = env_globals,
    expr = expr,
    extract = furrr_map2_extract
  )
}

furrr_map2_extract <- function(x, i) {
  map(x, `[`, i)
}

# ------------------------------------------------------------------------------

furrr_pmap_template <- function(
  l,
  fn,
  dots,
  options,
  progress,
  type,
  purrr_fn_name,
  env_globals
) {
  if (is.data.frame(l)) {
    l <- as.list(l)
  }

  n <- furrr_length_common(l)
  args <- furrr_recycle_common(l, n)

  if (length(l) == 0L) {
    # Special case handling of empty `l` like `future_pmap(list())`
    names <- NULL
  } else {
    # Get names after possible recycling
    names <- names(args[[1]])
  }

  assert_furrr_options(options)
  assert_progress(progress)

  progress <- reconcile_progress_with_strategy(progress)

  expr_seed_setup <- make_expr_seed_setup(options$seed)
  expr_seed_update <- make_expr_seed_update(options$seed)

  expr_progress_setup <- make_expr_progress_setup(progress)
  expr_progress_update <- make_expr_progress_update(progress)

  walk <- identical(purrr_fn_name, "pwalk")

  if (walk) {
    purrr_fn_name <- "pmap"
  }

  expr_out <- make_expr_out(walk)

  expr <- expr({
    ...furrr_chunk_l <- ...furrr_chunk_args

    !!expr_seed_setup
    !!expr_progress_setup

    ...furrr_fn_wrapper <- function(...) {
      !!expr_seed_update
      !!expr_progress_update

      ...furrr_out <- ...furrr_fn(...)

      !!expr_out
    }

    ...furrr_purrr_fn <- get(
      !!purrr_fn_name,
      envir = asNamespace("purrr"),
      mode = "function",
      inherits = FALSE
    )

    args <- list(
      .l = ...furrr_chunk_l,
      .f = ...furrr_fn_wrapper
    )

    args <- c(args, ...furrr_dots)

    do.call(...furrr_purrr_fn, args)
  })

  furrr_template(
    args = args,
    fn = fn,
    dots = dots,
    n = n,
    options = options,
    progress = progress,
    type = type,
    names = names,
    env_globals = env_globals,
    expr = expr,
    extract = furrr_pmap_extract
  )
}

furrr_pmap_extract <- function(x, i) {
  map(x, `[`, i)
}

# ------------------------------------------------------------------------------

furrr_template <- function(
  args,
  fn,
  dots,
  n,
  options,
  progress,
  type,
  names,
  env_globals,
  expr,
  extract
) {
  fn <- purrr::as_mapper(fn)

  if (is.null(options$seed) || is_false(options$seed)) {
    seeds <- NULL
  } else {
    seeds <- make_seeds(options$seed, n)

    # Ensure we step forward 1 seed on exit to ensure
    # cross-strategy reproducibility
    oseed <- next_random_seed()
    on.exit(set_random_seed(oseed), add = TRUE)

    # Pass `NULL` on to `future()` to signal that we shouldn't check for
    # RNG usage in the expression, as it is handled by furrr
    options$seed <- NULL
  }

  n_workers <- future::nbrOfWorkers()

  chunks <- make_chunks(n, n_workers, options$scheduling, options$chunk_size)
  n_chunks <- length(chunks)

  order <- make_order(n, options$scheduling, options$chunk_size)

  # Reorder `chunks` based on the custom `order`
  if (!is.null(order)) {
    chunks <- map(chunks, function(chunk) .subset(order, chunk))
  }

  gp <- get_globals_and_packages(
    options$globals,
    options$packages,
    fn,
    dots,
    env_globals
  )

  globals <- gp$globals
  packages <- gp$packages

  future_globals_max_size <- getOption("future.globals.maxSize")
  future_globals_max_size_default <- future_globals_max_size

  if (is.null(future_globals_max_size_default)) {
    future_globals_max_size_default <- 500 * 1024^2
  }

  # Expression adjustment for per worker max global size
  expr <- expr({
    future_globals_max_size <- getOption("future.globals.maxSize")

    if (!identical(future_globals_max_size, ...furrr_globals_max_size)) {
      options(future.globals.maxSize = ...furrr_globals_max_size)
      on.exit(
        options(future.globals.maxSize = future_globals_max_size),
        add = TRUE
      )
    }

    !!expr
  })

  if (is.null(options$prefix)) {
    labels <- NULL
  } else {
    labels <- paste0(options$prefix, "-", seq_len(n_chunks))
  }

  if (progress) {
    # nocov start
    objectSize <- import_future("objectSize")

    file <- tempfile(fileext = ".txt")

    file.create(file)
    on.exit(unlink(file, force = TRUE), add = TRUE)

    globals_file <- list(...furrr_progress_file = file)
    globals_file <- future::as.FutureGlobals(globals_file)
    globals_file <- future::resolve(globals_file)
    globals_file <- set_total_size(globals_file, objectSize(globals_file))

    globals <- c(globals, globals_file)
    # nocov end
  }

  scan_chunk_args_for_globals <- is_true(options$globals)

  futures <- vector("list", length = n_chunks)

  for (i in seq_len(n_chunks)) {
    chunk <- chunks[[i]]
    n_chunk <- length(chunk)

    chunk_args <- extract(args, chunk)

    chunk_globals <- globals
    chunk_globals[["...furrr_chunk_args"]] <- chunk_args

    chunk_packages <- packages

    if (scan_chunk_args_for_globals) {
      chunk_args_gp <- future::getGlobalsAndPackages(
        expr = chunk_args,
        envir = env_globals,
        globals = TRUE
      )

      chunk_args_globals <- chunk_args_gp$globals
      chunk_args_globals <- enforce_future_globals(chunk_args_globals)

      chunk_globals <- c(chunk_globals, chunk_args_globals)
      chunk_globals <- unique(chunk_globals)

      chunk_packages <- c(chunk_packages, chunk_args_gp$packages)
      chunk_packages <- unique(chunk_packages)
    }

    # Adjust "future.globals.maxSize" option to account for the fact that
    # more than one element is processed per future. Adjustment is done by
    # scaling up the limit by the number of elements in the current chunk.
    # This is an ad-hoc "good enough" approach, see:
    # https://github.com/HenrikBengtsson/future.apply/issues/8
    chunk_globals["...furrr_globals_max_size"] <- list(future_globals_max_size)
    options(future.globals.maxSize = n_chunk * future_globals_max_size_default)
    on.exit(
      options(future.globals.maxSize = future_globals_max_size),
      add = TRUE
    )

    if (!is.null(seeds)) {
      chunk_seeds <- seeds[chunk]
      chunk_globals[["...furrr_chunk_seeds"]] <- chunk_seeds
    }

    futures[[i]] <- future::future(
      expr,
      substitute = FALSE,
      envir = env_globals,
      stdout = options$stdout,
      conditions = options$conditions,
      globals = chunk_globals,
      packages = chunk_packages,
      seed = options$seed,
      label = labels[[i]]
    )
  }

  if (progress) {
    poll_progress(futures, file, n)
  }

  values <- furrr_try_catch({
    future::value(futures)
  })

  if (length(values) != length(chunks)) {
    abort("Internal error: Length of `values` not equal to length of `chunks`.")
  }

  out <- vctrs::vec_c(!!!values, .ptype = vector(type))

  if (!is.null(order)) {
    order_inv <- vector("integer", length = n)
    idx <- seq2(1L, n)
    order_inv[.subset(order, idx)] <- idx
    out <- out[order_inv]
  }

  names(out) <- names

  out
}

# ------------------------------------------------------------------------------

furrr_try_catch <- function(expr) {
  tryCatch(
    expr = expr,
    purrr_error_indexed = rethrow_purrr_error_indexed
  )
}

rethrow_purrr_error_indexed <- function(cnd) {
  # The purrr index information doesn't align with furrr's since the
  # input is chunked along the workers. We rethrow the parent instead,
  # which is the original error.
  cnd <- cnd$parent

  if (!is_condition(cnd)) {
    # If something changes in purrr and `$parent` doesn't exist, just
    # return and let normal erroring take its course
    return()
  }

  rlang::cnd_signal(cnd)
}

# ------------------------------------------------------------------------------

make_expr_seed_setup <- function(seed) {
  if (is.null(seed) || is_false(seed)) {
    return(NULL)
  }

  expr({
    ...furrr_chunk_seeds_env <- environment()
    ...furrr_chunk_seeds_env[["i"]] <- 1L
  })
}

make_expr_seed_update <- function(seed) {
  if (is.null(seed) || is_false(seed)) {
    return(NULL)
  }

  expr({
    ...furrr_chunk_seeds_i <- ...furrr_chunk_seeds_env[["i"]]
    ...furrr_chunk_seeds_env[["i"]] <- ...furrr_chunk_seeds_i + 1L

    assign(
      x = ".Random.seed",
      value = ...furrr_chunk_seeds[[...furrr_chunk_seeds_i]],
      envir = globalenv(),
      inherits = FALSE
    )
  })
}

# ------------------------------------------------------------------------------

# nocov start

make_expr_progress_setup <- function(progress) {
  if (is_false(progress)) {
    return(NULL)
  }

  expr({
    ...furrr_progress <- TRUE

    tryCatch(
      expr = {
        ...furrr_progress_con <- file(...furrr_progress_file, open = "a")
        on.exit(close(...furrr_progress_con), add = TRUE)
      },
      error = function(cnd) {
        ...furrr_progress <<- FALSE
      }
    )
  })
}

make_expr_progress_update <- function(progress) {
  if (is_false(progress)) {
    return(NULL)
  }

  expr({
    if (...furrr_progress) {
      try(
        expr = {
          cat("+", file = ...furrr_progress_con, sep = "")
        },
        silent = TRUE
      )
    }
  })
}

# nocov end

# ------------------------------------------------------------------------------

# nocov start

make_expr_out <- function(walk) {
  if (walk) {
    # If `walk/walk2/pwalk()` is being used, we internally switch it for its
    # `map()` equivalent and return `NULL` from each function call to avoid
    # sending either `.x` or the function result back to the main process.
    # We have to use the `map()` equivalent because `walk()` will force `.x` to
    # be returned back to the main process, even if each individual `.f` call
    # returns `NULL`. https://github.com/futureverse/furrr/issues/205
    expr(NULL)
  } else {
    expr(...furrr_out)
  }
}

# nocov end

# ------------------------------------------------------------------------------

# Required global variable hack for variables used in `expr()`.
# Required to pass R CMD check.
utils::globalVariables(
  c(
    "...furrr_out",
    "...furrr_chunk_args",
    "...furrr_fn",
    "...furrr_dots",
    "...furrr_globals_max_size",
    "...furrr_chunk_seeds",
    "...furrr_chunk_seeds_env",
    "...furrr_progress",
    "...furrr_progress_file",
    "...furrr_progress_con"
  )
)

#' Modify elements selectively via futures
#'
#' These functions work the same as [purrr::modify()] functions, but
#' allow you to modify in parallel.
#'
#' @param .x A vector.
#'
#' @param .f A function specified in the same way as the corresponding map
#' function.
#'
#' @inheritParams future_map_at
#'
#' @details
#' From purrr:
#'
#' Since the transformation can alter the structure of the input;
#' it's your responsibility to ensure that the transformation produces a valid
#' output. For example, if you're modifying a data frame, `.f` must preserve the
#' length of the input.
#'
#' @return An object the same class as `.x`
#'
#' @export
#' @examples
#' \donttest{plan(multisession, workers = 2)}
#'
#' # Convert each col to character, in parallel
#' future_modify(mtcars, as.character)
#'
#' iris |>
#'  future_modify_if(is.factor, as.character) |>
#'  str()
#'
#' mtcars |>
#'   future_modify_at(c(1, 4, 5), as.character) |>
#'   str()
#'
#' \dontshow{
#' # Close open connections for R CMD Check
#' if (!inherits(plan(), "sequential")) plan(sequential)
#' }
future_modify <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  UseMethod("future_modify")
}

#' @export
future_modify.default <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  result <- future_map(
    .x = .x,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  for (i in seq_along(.x)) {
    .x[[i]] <- result[[i]]
  }

  .x
}

#' @export
future_modify.character <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  result <- future_map_chr(
    .x = .x,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  .x[] <- result

  .x
}

#' @export
future_modify.double <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  result <- future_map_dbl(
    .x = .x,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  .x[] <- result

  .x
}

#' @export
future_modify.integer <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  result <- future_map_int(
    .x = .x,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  .x[] <- result

  .x
}

#' @export
future_modify.logical <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  result <- future_map_lgl(
    .x = .x,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  .x[] <- result

  .x
}

#' @export
future_modify.pairlist <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  result <- future_map(
    .x = .x,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  as.pairlist(result)
}

# ------------------------------------------------------------------------------

#' @rdname future_modify
#' @export
future_modify_at <- function(
  .x,
  .at,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  UseMethod("future_modify_at")
}

#' @export
future_modify_at.default <- function(
  .x,
  .at,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  where <- at_selection(names(.x), .at)
  sel <- inv_which(.x, where)

  future_modify_if(
    .x = .x,
    .p = sel,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )
}

#' @export
future_modify_at.integer <- function(
  .x,
  .at,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  where <- at_selection(names(.x), .at)
  sel <- inv_which(.x, where)

  .x[sel] <- future_map_int(
    .x = .x[sel],
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  .x
}

#' @export
future_modify_at.double <- function(
  .x,
  .at,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  where <- at_selection(names(.x), .at)
  sel <- inv_which(.x, where)

  .x[sel] <- future_map_dbl(
    .x = .x[sel],
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  .x
}

#' @export
future_modify_at.character <- function(
  .x,
  .at,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  where <- at_selection(names(.x), .at)
  sel <- inv_which(.x, where)

  .x[sel] <- future_map_chr(
    .x = .x[sel],
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  .x
}

#' @export
future_modify_at.logical <- function(
  .x,
  .at,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  where <- at_selection(names(.x), .at)
  sel <- inv_which(.x, where)

  .x[sel] <- future_map_lgl(
    .x = .x[sel],
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  .x
}

# ------------------------------------------------------------------------------

#' @rdname future_modify
#' @export
future_modify_if <- function(
  .x,
  .p,
  .f,
  ...,
  .else = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  UseMethod("future_modify_if")
}

#' @export
future_modify_if.default <- function(
  .x,
  .p,
  .f,
  ...,
  .else = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  sel <- probe(.x, .p)
  index <- seq_along(.x)

  result <- future_map(
    .x = .x[sel],
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  locs <- index[sel]

  for (i in seq_along(locs)) {
    loc <- locs[[i]]
    .x[[loc]] <- result[[i]]
  }

  if (!is_null(.else)) {
    result_else <- future_map(
      .x = .x[!sel],
      .f = .else,
      ...,
      .options = .options,
      .env_globals = .env_globals,
      .progress = FALSE
    )

    locs_else <- index[!sel]

    for (i in seq_along(locs_else)) {
      loc <- locs_else[[i]]
      .x[[loc]] <- result_else[[i]]
    }
  }

  .x
}

#' @export
future_modify_if.integer <- function(
  .x,
  .p,
  .f,
  ...,
  .else = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_modify_if_variant(
    .x = .x,
    .p = .p,
    .f = .f,
    ...,
    .else = .else,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress,
    .future_map_variant = future_map_int
  )
}

#' @export
future_modify_if.double <- function(
  .x,
  .p,
  .f,
  ...,
  .else = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_modify_if_variant(
    .x = .x,
    .p = .p,
    .f = .f,
    ...,
    .else = .else,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress,
    .future_map_variant = future_map_dbl
  )
}

#' @export
future_modify_if.character <- function(
  .x,
  .p,
  .f,
  ...,
  .else = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_modify_if_variant(
    .x = .x,
    .p = .p,
    .f = .f,
    ...,
    .else = .else,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress,
    .future_map_variant = future_map_chr
  )
}

#' @export
future_modify_if.logical <- function(
  .x,
  .p,
  .f,
  ...,
  .else = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_modify_if_variant(
    .x = .x,
    .p = .p,
    .f = .f,
    ...,
    .else = .else,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress,
    .future_map_variant = future_map_lgl
  )
}

future_modify_if_variant <- function(
  .x,
  .p,
  .f,
  ...,
  .else,
  .options,
  .env_globals,
  .progress,
  .future_map_variant
) {
  sel <- probe(.x, .p)

  result <- .future_map_variant(
    .x = .x[sel],
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  .x[sel] <- result

  if (!is_null(.else)) {
    not_sel <- !sel

    result <- .future_map_variant(
      .x = .x[not_sel],
      .f = .else,
      ...,
      .options = .options,
      .env_globals = .env_globals,
      .progress = FALSE
    )

    .x[not_sel] <- result
  }

  .x
}

# ------------------------------------------------------------------------------
# future_modify()

furrr_test_that("future_modify() default method works", {
  expect_identical(future_modify(list(1, 2), ~3), list(3, 3))
  expect_identical(
    future_modify(data.frame(x = 1, y = 2), ~3),
    data.frame(x = 3, y = 3)
  )
})

# TODO: Fix `NULL` behavior to match what is done here
# https://github.com/tidyverse/purrr/pull/754
furrr_test_that("future_modify() is not stable when returning `NULL`", {
  expect_identical(
    future_modify(
      list(1, 2),
      ~ if (.x == 1) {
        NULL
      } else {
        .x
      }
    ),
    list(2, 2)
  )
})

furrr_test_that("future_modify() variants work", {
  expect_identical(future_modify(c(1L, 2L, 3L), ~2L), rep(2L, 3))
  expect_identical(future_modify(c(1, 2, 3), ~2), rep(2, 3))
  expect_identical(future_modify(c("a", "b", "c"), toupper), c("A", "B", "C"))
  expect_identical(future_modify(c(TRUE, FALSE, TRUE), ~TRUE), rep(TRUE, 3))
})

furrr_test_that("future_modify(<pairlist>) works", {
  x <- as.pairlist(list(1, 2))
  expect_identical(class(future_modify(x, ~.x)), "pairlist")
})

# ------------------------------------------------------------------------------
# future_modify_at()

furrr_test_that("future_modify_at() default works", {
  expect_identical(future_modify_at(list(1, 2, 3), c(1, 3), ~5), list(5, 2, 5))
  expect_identical(
    future_modify_at(data.frame(x = 1, y = 2), 2, ~3),
    data.frame(x = 1, y = 3)
  )
})

furrr_test_that("future_modify_at() variants works", {
  expect_identical(future_modify_at(c(1L, 2L, 3L), c(1, 3), ~5L), c(5L, 2L, 5L))
  expect_identical(future_modify_at(c(1, 2, 3), c(1, 3), ~5), c(5, 2, 5))
  expect_identical(
    future_modify_at(c("a", "b", "c"), c(1, 3), toupper),
    c("A", "b", "C")
  )
  expect_identical(
    future_modify_at(c(TRUE, FALSE, TRUE), c(1, 3), ~NA),
    c(NA, FALSE, NA)
  )
})

# ------------------------------------------------------------------------------
# future_modify_if()

furrr_test_that("future_modify_if() default works", {
  expect_identical(future_modify_if(list(1, 2), ~ .x == 1, ~3), list(3, 2))
  expect_identical(
    future_modify_if(data.frame(x = 1, y = 2), ~ .x == 1, ~3),
    data.frame(x = 3, y = 2)
  )
})

furrr_test_that("future_modify_if() `.else` works for default", {
  expect_identical(
    future_modify_if(list(1, 2, 1, 4), ~ .x == 1, ~3, .else = ~4),
    list(3, 4, 3, 4)
  )
})

furrr_test_that("future_modify_if() variants works", {
  expect_identical(
    future_modify_if(c(1L, 2L, 3L), ~ .x == 1L, ~2L, .else = ~3L),
    c(2L, 3L, 3L)
  )
  expect_identical(
    future_modify_if(c(1, 2, 3), ~ .x == 1, ~2, .else = ~3),
    c(2, 3, 3)
  )
  expect_identical(
    future_modify_if(c("a", "b", "c"), ~ .x == "a", toupper, .else = ~"d"),
    c("A", "d", "d")
  )
  expect_identical(
    future_modify_if(c(TRUE, FALSE, TRUE), ~ .x == TRUE, ~TRUE, .else = ~NA),
    c(TRUE, NA, TRUE)
  )
})

#' Apply a function to each element of a vector, and its index via futures
#'
#' These functions work the same as [purrr::imap()] functions,
#' but allow you to map in parallel.
#'
#' @param .f A function, specified in one of the following ways:
#'
#'   * A named function, e.g. `paste`.
#'   * An anonymous function, e.g. `\(x, idx) x + idx` or
#'     `function(x, idx) x + idx`.
#'   * A formula, e.g. `~ .x + .y`. Use `.x` to refer to the current element and
#'     `.y` to refer to the current index. No longer recommended.
#'
#' @inheritParams future_map
#'
#' @return
#' A vector the same length as .x.
#'
#' @export
#' @examples
#' \donttest{plan(multisession, workers = 2)}
#'
#' future_imap_chr(sample(10), ~ paste0(.y, ": ", .x))
#'
#' \dontshow{
#' # Close open connections for R CMD Check
#' if (!inherits(plan(), "sequential")) plan(sequential)
#' }
future_imap <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_map2(
    .x = .x,
    .y = vec_index(.x),
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )
}

#' @rdname future_imap
#' @export
future_imap_chr <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_map2_chr(
    .x = .x,
    .y = vec_index(.x),
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )
}

#' @rdname future_imap
#' @export
future_imap_dbl <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_map2_dbl(
    .x = .x,
    .y = vec_index(.x),
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )
}

#' @rdname future_imap
#' @export
future_imap_int <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_map2_int(
    .x = .x,
    .y = vec_index(.x),
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )
}

#' @rdname future_imap
#' @export
future_imap_lgl <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_map2_lgl(
    .x = .x,
    .y = vec_index(.x),
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )
}

#' @rdname future_imap
#' @export
future_imap_vec <- function(
  .x,
  .f,
  ...,
  .ptype = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_map2_vec(
    .x = .x,
    .y = vec_index(.x),
    .f = .f,
    ...,
    .ptype = .ptype,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )
}

#' @rdname future_imap
#' @export
future_imap_dfr <- function(
  .x,
  .f,
  ...,
  .id = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_map2_dfr(
    .x = .x,
    .y = vec_index(.x),
    .f = .f,
    ...,
    .id = .id,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )
}

#' @rdname future_imap
#' @export
future_imap_dfc <- function(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  future_map2_dfc(
    .x = .x,
    .y = vec_index(.x),
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )
}

furrr_test_that("imap functions work with unnamed input", {
  expect_identical(future_imap(1:2, ~.y), list(1L, 2L))
  expect_identical(future_imap_chr(1:2, ~ as.character(.y)), c("1", "2"))
  expect_identical(future_imap_int(1:2, ~.y), c(1L, 2L))
  expect_identical(future_imap_dbl(1:2, ~.y), c(1, 2))
  expect_identical(future_imap_lgl(1:2, ~ identical(.y, 1L)), c(TRUE, FALSE))
  expect_identical(future_imap_vec(1:2, ~ identical(.y, 1L)), c(TRUE, FALSE))
  expect_identical(
    future_imap_dfr(1:2, ~ data.frame(x = .y)),
    data.frame(x = c(1L, 2L))
  )
  expect_identical(
    future_imap_dfc(1:2, ~ vctrs::new_data_frame(set_names(list(1), .y))),
    vctrs::new_data_frame(list(`1` = 1, `2` = 1))
  )
})

furrr_test_that("imap functions work with named input", {
  x <- set_names(1:2, c("x", "y"))
  expect_identical(future_imap(x, ~.y), list(x = "x", y = "y"))
  expect_identical(future_imap_chr(x, ~ as.character(.y)), c(x = "x", y = "y"))
  expect_identical(
    future_imap_int(x, ~ if (.y == "x") 1L else 2L),
    c(x = 1L, y = 2L)
  )
  expect_identical(
    future_imap_dbl(x, ~ if (.y == "x") 1 else 2),
    c(x = 1, y = 2)
  )
  expect_identical(
    future_imap_lgl(x, ~ if (.y == "x") TRUE else FALSE),
    c(x = TRUE, y = FALSE)
  )
  expect_identical(
    future_imap_vec(x, ~ if (.y == "x") TRUE else FALSE),
    c(x = TRUE, y = FALSE)
  )
  expect_identical(
    future_imap_dfr(x, ~ data.frame(x = .y)),
    data.frame(x = c("x", "y"))
  )
  expect_identical(
    future_imap_dfc(x, ~ vctrs::new_data_frame(set_names(list(1), .y))),
    vctrs::new_data_frame(list(x = 1, y = 1))
  )
})

# Detects leaked workers in tests, which show up as:
#
# ```
# Warning message:
# In .Internal(gc(verbose, reset, full)) :
#   closing unused connection 4 (<-localhost:11913)
# ```
#
# If a leaked connection is detected, a warning is thrown by testthat for
# the problematic test.
#
# We keep this on constantly to ensure we aren't leaking connections. It has
# also been helpful for detecting future issues around leaked workers (#307).
set_state_inspector(function() {
  getAllConnections()
})

get_globals_and_packages <- function(
  globals,
  packages,
  fn,
  dots,
  env_globals
) {
  objectSize <- import_future("objectSize")

  # Always need purrr on the worker
  packages_out <- "purrr"

  # Always get `.f`
  globals_fn <- list(...furrr_fn = fn)
  globals_fn <- future::as.FutureGlobals(globals_fn)
  globals_fn <- future::resolve(globals_fn)
  globals_fn <- set_total_size(globals_fn, objectSize(globals_fn))

  # Always get `...`.
  globals_dots <- list(...furrr_dots = dots)
  globals_dots <- future::as.FutureGlobals(globals_dots)
  globals_dots <- future::resolve(globals_dots)
  globals_dots <- set_total_size(globals_dots, objectSize(globals_dots))

  # Always get chunk specific placeholders
  globals_extra <- list(
    ...furrr_chunk_args = NULL,
    ...furrr_chunk_seeds = NULL,
    ...furrr_globals_max_size = NULL
  )
  globals_extra <- future::as.FutureGlobals(globals_extra)
  globals_extra <- future::resolve(globals_extra)
  globals_extra <- set_total_size(globals_extra, objectSize(globals_extra))

  globals_out <- c(
    globals_fn,
    globals_dots,
    globals_extra
  )

  # Collect all globals recursively
  # Search in the parent frame of the `future_*()` call for globals
  if (is_true(globals)) {
    # Lookup `.f` globals in the function env of `.f` (#153)
    env_fn <- fn_env(fn)

    gp_fn <- future::getGlobalsAndPackages(fn, envir = env_fn, globals = TRUE)
    gp_dots <- future::getGlobalsAndPackages(
      dots,
      envir = env_globals,
      globals = TRUE
    )

    globals_gp_fn <- enforce_future_globals(gp_fn$globals)
    globals_gp_dots <- enforce_future_globals(gp_dots$globals)

    globals_out <- unique(c(globals_out, globals_gp_fn, globals_gp_dots))
    packages_out <- unique(c(packages_out, gp_fn$packages, gp_dots$packages))
  }

  # Collect only explicitly selected globals,
  # but be lax about it with `mustExist = FALSE`.
  if (is.character(globals)) {
    globals_chr <- globals::globalsByName(
      globals,
      envir = env_globals,
      mustExist = FALSE
    )
    globals_chr <- enforce_future_globals(globals_chr)
    globals_out <- c(globals_out, globals_chr)
  }

  # Assume supplied named inputs are the globals.
  if (is.list(globals)) {
    globals_lst <- enforce_future_globals(globals)
    globals_out <- c(globals_out, globals_lst)
  }

  # Add user specified packages
  if (!is.null(packages)) {
    packages_out <- unique(c(packages_out, packages))
  }

  out <- list(globals = globals_out, packages = packages_out)

  out
}

# ------------------------------------------------------------------------------

# `getGlobalsAndPackages()` should always return a FutureGlobals object
# with a non-NA `total_size`, but HenrikBengtsson/future#410 proves it
# doesn't, so be careful and ensure correct behavior
enforce_future_globals <- function(x) {
  x <- future::as.FutureGlobals(x)

  if (is.na(get_total_size(x))) {
    objectSize <- import_future("objectSize")
    x <- set_total_size(x, objectSize(x))
  }

  x
}

#' @keywords internal
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
#' @import rlang
#' @importFrom purrr map
#' @importFrom future future
#' @importFrom parallel nextRNGSubStream
#' @importFrom parallel nextRNGStream
## usethis namespace: end
NULL

#' @rdname future_map2
#' @export
future_pmap <- function(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_pmap_template(
    l = .l,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "list",
    purrr_fn_name = "pmap",
    env_globals = .env_globals
  )
}

#' @rdname future_map2
#' @export
future_pmap_chr <- function(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_pmap_template(
    l = .l,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "character",
    purrr_fn_name = "pmap_chr",
    env_globals = .env_globals
  )
}

#' @rdname future_map2
#' @export
future_pmap_dbl <- function(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_pmap_template(
    l = .l,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "double",
    purrr_fn_name = "pmap_dbl",
    env_globals = .env_globals
  )
}

#' @rdname future_map2
#' @export
future_pmap_int <- function(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_pmap_template(
    l = .l,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "integer",
    purrr_fn_name = "pmap_int",
    env_globals = .env_globals
  )
}

#' @rdname future_map2
#' @export
future_pmap_lgl <- function(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  furrr_pmap_template(
    l = .l,
    fn = .f,
    dots = list(...),
    options = .options,
    progress = .progress,
    type = "logical",
    purrr_fn_name = "pmap_lgl",
    env_globals = .env_globals
  )
}

#' @rdname future_map2
#' @export
future_pmap_vec <- function(
  .l,
  .f,
  ...,
  .ptype = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  out <- future_pmap(
    .l = .l,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  simplify_impl(
    out,
    ptype = .ptype,
    error_arg = "<output>",
    error_call = current_env()
  )
}

#' @rdname future_map2
#' @export
future_pmap_dfr <- function(
  .l,
  .f,
  ...,
  .id = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  if (!rlang::is_installed("dplyr")) {
    rlang::abort("`future_map_dfr()` requires dplyr")
  }

  res <- future_pmap(
    .l = .l,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  dplyr::bind_rows(res, .id = .id)
}

#' @rdname future_map2
#' @export
future_pmap_dfc <- function(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
) {
  if (!rlang::is_installed("dplyr")) {
    rlang::abort("`future_map_dfc()` requires dplyr")
  }

  res <- future_pmap(
    .l = .l,
    .f = .f,
    ...,
    .options = .options,
    .env_globals = .env_globals,
    .progress = .progress
  )

  dplyr::bind_cols(res)
}

# ------------------------------------------------------------------------------
# pmap()

furrr_test_that("future_pmap() matches pmap() for simple cases", {
  expect_identical(
    future_pmap(list(1:3, 4:6, 7:9), ~ .x + .y + ..3),
    pmap(list(1:3, 4:6, 7:9), ~ .x + .y + ..3)
  )
})

furrr_test_that("names of `.x` are retained", {
  x <- c(a = 1, b = 2)
  y <- c(c = 1, d = 2)
  expect_named(future_pmap(list(x, y), ~1), c("a", "b"))
})

furrr_test_that("named empty input makes named empty output", {
  x <- set_names(list(), character())
  expect_named(future_pmap(list(x, x), ~.x), character())
})

# ------------------------------------------------------------------------------
# atomic variants

furrr_test_that("future_pmap_dbl() works", {
  x <- c(1, 2, 3)
  y <- c(4, 5, 6)

  expect_identical(
    future_pmap_dbl(list(x, y), ~ .x + .y),
    pmap_dbl(list(x, y), ~ .x + .y)
  )
})

furrr_test_that("future_pmap_int() works", {
  x <- c(1L, 2L, 3L)
  y <- c(4L, 5L, 6L)

  expect_identical(
    future_pmap_int(list(x, y), ~ .x + .y),
    pmap_int(list(x, y), ~ .x + .y)
  )
})

furrr_test_that("future_pmap_lgl() works", {
  x <- c(TRUE, FALSE, TRUE)
  y <- c(FALSE, TRUE, TRUE)

  expect_identical(
    future_pmap_lgl(list(x, y), ~ .x || .y),
    pmap_lgl(list(x, y), ~ .x || .y)
  )
})

furrr_test_that("future_pmap_chr() works", {
  x <- c("a", "b", "c")
  y <- c("d", "e", "f")

  expect_identical(
    future_pmap_chr(list(x, y), ~.y),
    pmap_chr(list(x, y), ~.y)
  )
})

furrr_test_that("future_pmap_vec() works", {
  x <- as.Date(c("2020-01-01", "2020-01-02", "2020-01-03"))
  y <- 1:3

  expect_identical(
    future_pmap_vec(list(x, y), ~.x),
    pmap_vec(list(x, y), ~.x)
  )
  expect_identical(
    future_pmap_vec(list(x, y), ~1, .ptype = integer()),
    pmap_vec(list(x, y), ~1, .ptype = integer())
  )

  expect_identical(
    future_pmap_vec(list(integer(), integer()), identity),
    pmap_vec(list(integer(), integer()), identity)
  )
  expect_identical(
    future_pmap_vec(
      list(set_names(integer(), character()), integer()),
      identity
    ),
    pmap_vec(
      list(set_names(integer(), character()), integer()),
      identity
    )
  )

  # Vector error
  expect_snapshot(error = TRUE, {
    future_pmap_vec(list(1:2, 1:2), ~NULL)
  })
  # Size error
  expect_snapshot(error = TRUE, {
    future_pmap_vec(list(1:2, 1:2), ~ 1:2)
  })
  # Type error
  expect_snapshot(error = TRUE, {
    future_pmap_vec(list(1:2, 1:2), ~ if (.x == 1L) 1 else "x")
  })
})

furrr_test_that("names of `.x` are retained", {
  x <- c(a = 1, b = 2)
  y <- c(c = 1, d = 2)
  expect_named(future_pmap_dbl(list(x, y), ~1), c("a", "b"))

  x <- c(a = as.Date("2020-01-01"), b = as.Date("2020-01-02"))
  y <- c(c = 1, d = 2)
  expect_named(future_pmap_vec(list(x, y), ~1), c("a", "b"))
})

# ------------------------------------------------------------------------------
# data frame variants

furrr_test_that("future_pmap_dfr() works", {
  x <- c("a", "b", "c")
  y <- c("d", "e", "f")

  expect_identical(
    future_pmap_dfr(list(x, y), ~ data.frame(x = .x, y = .y)),
    pmap_dfr(list(x, y), ~ data.frame(x = .x, y = .y))
  )
})

furrr_test_that("future_pmap_dfc() works", {
  x <- c("a", "b", "c")
  y <- c("d", "e", "f")

  expect_identical(
    future_pmap_dfc(list(x, y), ~ as.data.frame(set_names(list(.x), .y))),
    pmap_dfc(list(x, y), ~ as.data.frame(set_names(list(.x), .y)))
  )
})

# ------------------------------------------------------------------------------
# size

furrr_test_that("future_pmap() works with completely empty list", {
  expect_identical(future_pmap(list(), identity), list())
  expect_identical(future_pmap_dbl(list(), identity), double())
})

furrr_test_that("future_pmap() works with size zero input", {
  expect_identical(future_pmap(list(list(), list()), identity), list())
})

furrr_test_that("atomic variants work with size zero input", {
  expect_identical(future_pmap_chr(list(list(), list()), identity), character())
  expect_identical(future_pmap_dbl(list(list(), list()), identity), double())
  expect_identical(future_pmap_int(list(list(), list()), identity), integer())
  expect_identical(future_pmap_lgl(list(list(), list()), identity), logical())
})

furrr_test_that("generic variant works with size zero input", {
  expect_identical(
    future_pmap_vec(list(list(), list()), identity),
    NULL
  )
  expect_identical(
    future_pmap_vec(list(list(), list()), identity, .ptype = integer()),
    integer()
  )
})

furrr_test_that("size one recycling works", {
  expect_identical(
    future_pmap(list(1, 1:2), ~ c(.x, .y)),
    list(c(1, 1), c(1, 2))
  )

  expect_identical(
    future_pmap(list(1:2, 1), ~ c(.x, .y)),
    list(c(1, 1), c(2, 1))
  )

  expect_identical(
    future_pmap(list(integer(), 1), ~ c(.x, .y)),
    list()
  )

  expect_identical(
    future_pmap(list(1, integer()), ~ c(.x, .y)),
    list()
  )
})

# TODO: Reenable this test after future issue is fixed
# https://github.com/futureverse/future/issues/820
# https://github.com/futureverse/furrr/issues/307
# furrr_test_that("generally can't recycle to size zero", {
#   expect_error(
#     future_pmap(list(1:2, integer()), ~ c(.x, .y)),
#     "Can't recycle"
#   )
#
#   expect_error(
#     future_pmap(list(integer(), 1:2), ~ c(.x, .y)),
#     "Can't recycle"
#   )
# })

# ------------------------------------------------------------------------------
# Miscellaneous

furrr_test_that("named arguments can be passed through", {
  vec_mean <- function(.x, .y, na.rm = FALSE) {
    mean(c(.x, .y), na.rm = na.rm)
  }

  x <- list(c(NA, 1), 1:2)

  expect_identical(
    future_pmap(x, vec_mean, na.rm = TRUE),
    list(1, 1.5)
  )
})

furrr_test_that("arguments can be matched by name", {
  x <- list(x = c(1, 2), y = c(3, 5))

  fn <- function(y, x) {
    y - x
  }

  expect_identical(future_pmap_dbl(x, fn), c(2, 3))
})

furrr_test_that("unused components can be absorbed", {
  x <- list(c(1, 2), c(3, 5))

  fn1 <- function(x) {
    x
  }
  fn2 <- function(x, ...) {
    x
  }

  # TODO: Reenable this test after future issue is fixed
  # https://github.com/futureverse/future/issues/820
  # https://github.com/futureverse/furrr/issues/307
  # expect_error(future_pmap_dbl(x, fn1))

  expect_identical(future_pmap_dbl(x, fn2), c(1, 2))
})

furrr_test_that("globals in `.x` and `.y` are found (#16)", {
  fn1 <- function(x) sum(x, na.rm = TRUE)
  fn2 <- function(x) sum(x, na.rm = FALSE)

  x <- list(c(1, 2, NA), c(2, 3, 4))

  fns1 <- map(x, ~ purrr::partial(fn1, x = .x))
  fns2 <- map(x, ~ purrr::partial(fn2, x = .x))

  expect_identical(
    future_pmap(list(fns1, fns2), ~ c(.x(), .y())),
    list(c(3, NA), c(9, 9))
  )
})

test_that("globals in `.l` are only exported to workers that use them", {
  plan(multisession, workers = 2)
  on.exit(plan(sequential), add = TRUE)

  # Use `local()` to ensure that the wrapper functions and the anonymous
  # functions created with `~` don't pick up extra globals
  my_wrapper1 <- local({
    my_mean1 <- function(x) mean(x, na.rm = TRUE)

    function(x) {
      my_mean1(x)
      exists("my_mean1")
    }
  })

  my_wrapper2 <- local({
    my_mean2 <- function(x) mean(x, na.rm = FALSE)

    function(x) {
      my_mean2(x)
      exists("my_mean1")
    }
  })

  x <- list(my_wrapper1, my_wrapper2)

  expect_identical(
    future_pmap_lgl(list(x), .f = ~ .x(c(1, NA))),
    c(TRUE, FALSE)
  )
})

furrr_test_that <- function(desc, code) {
  test_code <- enquo(code)

  # Ensure that we reset to sequential plan even on error to
  # close any straggler processes
  on.exit(plan(sequential), add = TRUE)

  for (strategy in supported_strategies()) {
    for (cores in seq2(1L, supported_max_cores(strategy))) {
      if (identical(strategy, "sequential")) {
        plan(strategy)
      } else {
        plan(strategy, workers = cores)
      }

      test_desc <- paste0(
        desc,
        " / strategy - ",
        strategy,
        " / cores - ",
        cores
      )
      test_expr <- expr(
        test_that(test_desc, {
          !!test_code
        })
      )

      eval_tidy(test_expr)
    }
  }
}

supported_strategies <- function() {
  strategies <- c("sequential", "multisession", "multicore")

  if (!parallelly::supportsMulticore()) {
    strategies <- setdiff(strategies, "multicore")
  }

  strategies
}

supported_max_cores <- function(strategy) {
  if (identical(strategy, "sequential")) {
    1L
  } else {
    parallelly::availableCores(max = 2L)
  }
}

# can use `stdout = NA` to not intercept at all

    Code
      future_map(1:2, fn, .options = opts)
    Output
      hello
      hello
      [[1]]
      [1] 1
      
      [[2]]
      [1] 2
      

---

    Code
      future_map(1:2, fn, .options = opts)
    Output
      [[1]]
      [1] 1
      
      [[2]]
      [1] 2
      

# validates `stdout`

    Code
      (expect_error(furrr_options(stdout = "x")))
    Output
      <error/rlang_error>
      Error in `validate_stdout()`:
      ! `stdout` must be `TRUE`, `FALSE`, or `NA`.
    Code
      (expect_error(furrr_options(stdout = c(TRUE, TRUE))))
    Output
      <error/vctrs_error_assert_size>
      Error in `validate_stdout()`:
      ! `stdout` must have size 1, not size 2.

# can avoid handling conditions altogether

    Code
      future_map(1:5, fn, .options = opts)
    Condition
      Warning in `...furrr_fn()`:
      hello
      Warning in `...furrr_fn()`:
      hello
      Warning in `...furrr_fn()`:
      hello
      Warning in `...furrr_fn()`:
      hello
      Warning in `...furrr_fn()`:
      hello
    Output
      [[1]]
      [1] 1
      
      [[2]]
      [1] 2
      
      [[3]]
      [1] 3
      
      [[4]]
      [1] 4
      
      [[5]]
      [1] 5
      

---

    Code
      future_map(1:5, fn, .options = opts)
    Output
      [[1]]
      [1] 1
      
      [[2]]
      [1] 2
      
      [[3]]
      [1] 3
      
      [[4]]
      [1] 4
      
      [[5]]
      [1] 5
      

# validates `conditions`

    Code
      (expect_error(furrr_options(conditions = 1)))
    Output
      <error/rlang_error>
      Error in `validate_conditions()`:
      ! `conditions` must be a character vector.

# future_pmap_vec() works / strategy - sequential / cores - 1

    Code
      future_pmap_vec(list(1:2, 1:2), ~NULL)
    Condition
      Error in `future_pmap_vec()`:
      ! `<output>[[1]]` must be a vector, not `NULL`.
      i Read our FAQ about scalar types (`?vctrs::faq_error_scalar_type`) to learn more.

---

    Code
      future_pmap_vec(list(1:2, 1:2), ~ 1:2)
    Condition
      Error in `future_pmap_vec()`:
      ! `<output>[[1]]` must have size 1, not size 2.

---

    Code
      future_pmap_vec(list(1:2, 1:2), ~ if (.x == 1L) 1 else "x")
    Condition
      Error in `future_pmap_vec()`:
      ! Can't combine `<output>[[1]]` <double> and `<output>[[2]]` <character>.

# future_pmap_vec() works / strategy - multisession / cores - 1

    Code
      future_pmap_vec(list(1:2, 1:2), ~NULL)
    Condition
      Error in `future_pmap_vec()`:
      ! `<output>[[1]]` must be a vector, not `NULL`.
      i Read our FAQ about scalar types (`?vctrs::faq_error_scalar_type`) to learn more.

---

    Code
      future_pmap_vec(list(1:2, 1:2), ~ 1:2)
    Condition
      Error in `future_pmap_vec()`:
      ! `<output>[[1]]` must have size 1, not size 2.

---

    Code
      future_pmap_vec(list(1:2, 1:2), ~ if (.x == 1L) 1 else "x")
    Condition
      Error in `future_pmap_vec()`:
      ! Can't combine `<output>[[1]]` <double> and `<output>[[2]]` <character>.

# future_pmap_vec() works / strategy - multisession / cores - 2

    Code
      future_pmap_vec(list(1:2, 1:2), ~NULL)
    Condition
      Error in `future_pmap_vec()`:
      ! `<output>[[1]]` must be a vector, not `NULL`.
      i Read our FAQ about scalar types (`?vctrs::faq_error_scalar_type`) to learn more.

---

    Code
      future_pmap_vec(list(1:2, 1:2), ~ 1:2)
    Condition
      Error in `future_pmap_vec()`:
      ! `<output>[[1]]` must have size 1, not size 2.

---

    Code
      future_pmap_vec(list(1:2, 1:2), ~ if (.x == 1L) 1 else "x")
    Condition
      Error in `future_pmap_vec()`:
      ! Can't combine `<output>[[1]]` <double> and `<output>[[2]]` <character>.

# future_pmap_vec() works / strategy - multicore / cores - 1

    Code
      future_pmap_vec(list(1:2, 1:2), ~NULL)
    Condition
      Error in `future_pmap_vec()`:
      ! `<output>[[1]]` must be a vector, not `NULL`.
      i Read our FAQ about scalar types (`?vctrs::faq_error_scalar_type`) to learn more.

---

    Code
      future_pmap_vec(list(1:2, 1:2), ~ 1:2)
    Condition
      Error in `future_pmap_vec()`:
      ! `<output>[[1]]` must have size 1, not size 2.

---

    Code
      future_pmap_vec(list(1:2, 1:2), ~ if (.x == 1L) 1 else "x")
    Condition
      Error in `future_pmap_vec()`:
      ! Can't combine `<output>[[1]]` <double> and `<output>[[2]]` <character>.

# future_pmap_vec() works / strategy - multicore / cores - 2

    Code
      future_pmap_vec(list(1:2, 1:2), ~NULL)
    Condition
      Error in `future_pmap_vec()`:
      ! `<output>[[1]]` must be a vector, not `NULL`.
      i Read our FAQ about scalar types (`?vctrs::faq_error_scalar_type`) to learn more.

---

    Code
      future_pmap_vec(list(1:2, 1:2), ~ 1:2)
    Condition
      Error in `future_pmap_vec()`:
      ! `<output>[[1]]` must have size 1, not size 2.

---

    Code
      future_pmap_vec(list(1:2, 1:2), ~ if (.x == 1L) 1 else "x")
    Condition
      Error in `future_pmap_vec()`:
      ! Can't combine `<output>[[1]]` <double> and `<output>[[2]]` <character>.

# future_map_vec() works / strategy - sequential / cores - 1

    Code
      future_map_vec(1:2, ~NULL)
    Condition
      Error in `future_map_vec()`:
      ! `<output>[[1]]` must be a vector, not `NULL`.
      i Read our FAQ about scalar types (`?vctrs::faq_error_scalar_type`) to learn more.

---

    Code
      future_map_vec(1:2, ~ 1:2)
    Condition
      Error in `future_map_vec()`:
      ! `<output>[[1]]` must have size 1, not size 2.

---

    Code
      future_map_vec(1:2, ~ if (.x == 1L) 1 else "x")
    Condition
      Error in `future_map_vec()`:
      ! Can't combine `<output>[[1]]` <double> and `<output>[[2]]` <character>.

# future_map_vec() works / strategy - multisession / cores - 1

    Code
      future_map_vec(1:2, ~NULL)
    Condition
      Error in `future_map_vec()`:
      ! `<output>[[1]]` must be a vector, not `NULL`.
      i Read our FAQ about scalar types (`?vctrs::faq_error_scalar_type`) to learn more.

---

    Code
      future_map_vec(1:2, ~ 1:2)
    Condition
      Error in `future_map_vec()`:
      ! `<output>[[1]]` must have size 1, not size 2.

---

    Code
      future_map_vec(1:2, ~ if (.x == 1L) 1 else "x")
    Condition
      Error in `future_map_vec()`:
      ! Can't combine `<output>[[1]]` <double> and `<output>[[2]]` <character>.

# future_map_vec() works / strategy - multisession / cores - 2

    Code
      future_map_vec(1:2, ~NULL)
    Condition
      Error in `future_map_vec()`:
      ! `<output>[[1]]` must be a vector, not `NULL`.
      i Read our FAQ about scalar types (`?vctrs::faq_error_scalar_type`) to learn more.

---

    Code
      future_map_vec(1:2, ~ 1:2)
    Condition
      Error in `future_map_vec()`:
      ! `<output>[[1]]` must have size 1, not size 2.

---

    Code
      future_map_vec(1:2, ~ if (.x == 1L) 1 else "x")
    Condition
      Error in `future_map_vec()`:
      ! Can't combine `<output>[[1]]` <double> and `<output>[[2]]` <character>.

# future_map_vec() works / strategy - multicore / cores - 1

    Code
      future_map_vec(1:2, ~NULL)
    Condition
      Error in `future_map_vec()`:
      ! `<output>[[1]]` must be a vector, not `NULL`.
      i Read our FAQ about scalar types (`?vctrs::faq_error_scalar_type`) to learn more.

---

    Code
      future_map_vec(1:2, ~ 1:2)
    Condition
      Error in `future_map_vec()`:
      ! `<output>[[1]]` must have size 1, not size 2.

---

    Code
      future_map_vec(1:2, ~ if (.x == 1L) 1 else "x")
    Condition
      Error in `future_map_vec()`:
      ! Can't combine `<output>[[1]]` <double> and `<output>[[2]]` <character>.

# future_map_vec() works / strategy - multicore / cores - 2

    Code
      future_map_vec(1:2, ~NULL)
    Condition
      Error in `future_map_vec()`:
      ! `<output>[[1]]` must be a vector, not `NULL`.
      i Read our FAQ about scalar types (`?vctrs::faq_error_scalar_type`) to learn more.

---

    Code
      future_map_vec(1:2, ~ 1:2)
    Condition
      Error in `future_map_vec()`:
      ! `<output>[[1]]` must have size 1, not size 2.

---

    Code
      future_map_vec(1:2, ~ if (.x == 1L) 1 else "x")
    Condition
      Error in `future_map_vec()`:
      ! Can't combine `<output>[[1]]` <double> and `<output>[[2]]` <character>.

# errors don't report purrr's indices (#250) / strategy - sequential / cores - 1

    Code
      future_map(x, fail_on_five)
    Condition
      Error in `...furrr_fn()`:
      ! Failure!

# errors don't report purrr's indices (#250) / strategy - multisession / cores - 1

    Code
      future_map(x, fail_on_five)
    Condition
      Error in `...furrr_fn()`:
      ! Failure!

# errors don't report purrr's indices (#250) / strategy - multisession / cores - 2

    Code
      future_map(x, fail_on_five)
    Condition
      Error in `...furrr_fn()`:
      ! Failure!

# errors don't report purrr's indices (#250) / strategy - multicore / cores - 1

    Code
      future_map(x, fail_on_five)
    Condition
      Error in `...furrr_fn()`:
      ! Failure!

# errors don't report purrr's indices (#250) / strategy - multicore / cores - 2

    Code
      future_map(x, fail_on_five)
    Condition
      Error in `...furrr_fn()`:
      ! Failure!

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-map.R, R/future-walk.R
\name{future_map}
\alias{future_map}
\alias{future_map_chr}
\alias{future_map_dbl}
\alias{future_map_int}
\alias{future_map_lgl}
\alias{future_map_vec}
\alias{future_map_dfr}
\alias{future_map_dfc}
\alias{future_walk}
\title{Apply a function to each element of a vector via futures}
\usage{
future_map(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map_chr(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map_dbl(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map_int(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map_lgl(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map_vec(
  .x,
  .f,
  ...,
  .ptype = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map_dfr(
  .x,
  .f,
  ...,
  .id = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map_dfc(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_walk(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)
}
\arguments{
\item{.x}{A list or atomic vector.}

\item{.f}{A function, specified in one of the following ways:
\itemize{
\item A named function, e.g. \code{mean}.
\item An anonymous function, e.g. \verb{\\(x) x + 1} or \code{function(x) x + 1}.
\item A formula, e.g. \code{~ .x + 1}. Use \code{.x} to refer to the first
argument. No longer recommended.
\item A string, integer, or list, e.g. \code{"idx"}, \code{1}, or \code{list("idx", 1)} which
are shorthand for \verb{\\(x) pluck(x, "idx")}, \verb{\\(x) pluck(x, 1)}, and
\verb{\\(x) pluck(x, "idx", 1)} respectively. Optionally supply \code{.default} to
set a default value if the indexed element is \code{NULL} or does not exist.
}}

\item{...}{Additional arguments passed on to the mapped function.

We now generally recommend against using \code{...} to pass additional
(constant) arguments to \code{.f}. Instead use a shorthand anonymous function:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{# Instead of
x |> future_map(f, 1, 2, collapse = ",")
# do:
x |> future_map(\\(x) f(x, 1, 2, collapse = ","))
}\if{html}{\out{</div>}}

This makes it easier to understand which arguments belong to which
function and will tend to yield better error messages.}

\item{.options}{The \code{future} specific options to use with the workers. This
must be the result from a call to \code{\link[=furrr_options]{furrr_options()}}.}

\item{.env_globals}{The environment to look for globals required by \code{.x} and
\code{...}. Globals required by \code{.f} are looked up in the function environment
of \code{.f}.}

\item{.progress}{A single logical. Should a progress bar be displayed?
Only works with multisession, multicore, and multiprocess futures. Note
that if a multicore/multisession future falls back to sequential, then
a progress bar will not be displayed.

\strong{Warning:} The \code{.progress} argument will be deprecated and removed
in a future version of furrr in favor of using the more robust
\href{https://CRAN.R-project.org/package=progressr}{progressr}
package.}

\item{.ptype}{If \code{NULL}, the default, the output type is the common type of
the elements of the result. Otherwise, supply a "prototype" giving the
desired type of output.}

\item{.id}{Either a string or \code{NULL}. If a string, the output will contain
a variable with that name, storing either the name (if \code{.x} is named) or
the index (if \code{.x} is unnamed) of the input. If \code{NULL}, the default, no
variable will be created.

Only applies to \verb{_dfr} variant.}
}
\value{
All functions return a vector the same length as \code{.x}.
\itemize{
\item \code{\link[=future_map]{future_map()}} returns a list
\item \code{\link[=future_map_lgl]{future_map_lgl()}} a logical vector
\item \code{\link[=future_map_int]{future_map_int()}} an integer vector
\item \code{\link[=future_map_dbl]{future_map_dbl()}} a double vector
\item \code{\link[=future_map_chr]{future_map_chr()}} a character vector
}

The output of \code{.f} will be automatically typed upwards, e.g. logical ->
integer -> double -> character.
}
\description{
These functions work the same as \code{\link[purrr:map]{purrr::map()}} and its variants, but
allow you to map in parallel.
}
\examples{
\donttest{plan(multisession, workers = 2)}

1:10 |>
  future_map(rnorm, n = 10, .options = furrr_options(seed = 123)) |>
  future_map_dbl(mean)

# If each element of the output is a data frame, use
# `future_map_dfr()` to row-bind them together:
mtcars |>
  split(mtcars$cyl) |>
  future_map(~ lm(mpg ~ wt, data = .x)) |>
  future_map_dfr(~ as.data.frame(t(as.matrix(coef(.)))))


# You can be explicit about what gets exported to the workers.
# To see this, use multisession (not multicore as the forked workers
# still have access to this environment)
\donttest{plan(multisession)}
x <- 1
y <- 2

# This will fail, y is not exported (no black magic occurs)
try(future_map(1, ~y, .options = furrr_options(globals = "x")))

# y is exported
future_map(1, ~y, .options = furrr_options(globals = "y"))

\dontshow{
# Close open connections for R CMD Check
if (!inherits(plan(), "sequential")) plan(sequential)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/furrr-options.R
\name{furrr_options}
\alias{furrr_options}
\title{Options to fine tune furrr}
\usage{
furrr_options(
  ...,
  stdout = TRUE,
  conditions = "condition",
  globals = TRUE,
  packages = NULL,
  seed = FALSE,
  scheduling = 1,
  chunk_size = NULL,
  prefix = NULL
)
}
\arguments{
\item{...}{These dots are reserved for future extensibility and must
be empty.}

\item{stdout}{A logical.
\itemize{
\item If \code{TRUE}, standard output of the underlying futures is relayed as soon
as possible.
\item If \code{FALSE}, output is silenced by sinking it to the null device.
}}

\item{conditions}{A character string of conditions classes to be relayed.
The default is to relay all conditions, including messages and warnings.
Errors are always relayed. To not relay any conditions (besides errors),
use \code{conditions = character()}. To selectively ignore specific classes,
use \code{conditions = structure("condition", exclude = "message")}.}

\item{globals}{A logical, a character vector, a named list, or \code{NULL} for
controlling how globals are handled. For details, see the
\verb{Global variables} section below.}

\item{packages}{A character vector, or \code{NULL}. If supplied, this specifies
packages that are guaranteed to be attached in the R environment where the
future is evaluated.}

\item{seed}{A logical, an integer of length \code{1} or \code{7}, a list of
\code{length(.x)} with pre-generated random seeds, or \code{NULL}. For details, see
the \verb{Reproducible random number generation (RNG)} section below.}

\item{scheduling}{A single integer, logical, or \code{Inf}. This argument
controls the average number of futures ("chunks") per worker.
\itemize{
\item If \code{0}, then a single future is used to process all elements of \code{.x}.
\item If \code{1} or \code{TRUE}, then one future per worker is used.
\item If \code{2}, then each worker will process two futures (provided there
are enough elements in \code{.x}).
\item If \code{Inf} or \code{FALSE}, then one future per element of \code{.x} is used.
}

This argument is only used if \code{chunk_size} is \code{NULL}.}

\item{chunk_size}{A single integer, \code{Inf}, or \code{NULL}. This argument
controls the average number of elements per future (\code{"chunk"}). If \code{Inf},
then all elements are processed in a single future. If \code{NULL}, then
\code{scheduling} is used instead to determine how \code{.x} is chunked.}

\item{prefix}{A single character string, or \code{NULL}. If a character string,
then each future is assigned a label as \code{{prefix}-{chunk-id}}. If \code{NULL},
no labels are used.}
}
\description{
\code{furrr_options()} returns an object that can be supplied as the \code{.options}
argument for furrr functions, such as \code{\link[=future_map]{future_map()}}. The options are either
used by furrr directly, or are passed on to \code{\link[future:future]{future::future()}}.
}
\section{Global variables}{


\code{globals} controls how globals are identified, similar to the \code{globals}
argument of \code{\link[future:future]{future::future()}}. Since all function calls use the same set of
globals, furrr gathers globals upfront (once), which is more efficient than
if it was done for each future independently.
\itemize{
\item If \code{TRUE} or \code{NULL}, then globals are automatically identified and
gathered.
\item If a character vector of names is specified, then those globals are
gathered.
\item If a named list, then those globals are used as is.
\item In all cases, \code{.f} and any \code{...} arguments are automatically passed as
globals to each future created, as they are always needed.
}
}

\section{Reproducible random number generation (RNG)}{


Unless \code{seed = FALSE}, furrr functions are guaranteed to generate
the exact same sequence of random numbers \emph{given the same initial
seed / RNG state} regardless of the type of futures and scheduling
("chunking") strategy.

Setting \code{seed = NULL} is equivalent to \code{seed = FALSE}, except that the
\code{future.rng.onMisuse} option is not consulted to potentially monitor the
future for faulty random number usage. See the \code{seed} argument of
\code{\link[future:future]{future::future()}} for more details.

RNG reproducibility is achieved by pre-generating the random seeds for all
iterations (over \code{.x}) by using L'Ecuyer-CMRG RNG streams. In each
iteration, these seeds are set before calling \code{.f(.x[[i]], ...)}.
\emph{Note, for large \code{length(.x)} this may introduce a large overhead.}

A fixed \code{seed} may be given as an integer vector, either as a full
L'Ecuyer-CMRG RNG seed of length \code{7}, or as a seed of length \code{1} that
will be used to generate a full L'Ecuyer-CMRG seed.

If \code{seed = TRUE}, then \code{.Random.seed} is returned if it holds a
L'Ecuyer-CMRG RNG seed, otherwise one is created randomly.

If \code{seed = NA}, a L'Ecuyer-CMRG RNG seed is randomly created.

If none of the function calls \code{.f(.x[[i]], ...)} use random number
generation, then \code{seed = FALSE} may be used.

In addition to the above, it is possible to specify a pre-generated
sequence of RNG seeds as a list such that \code{length(seed) == length(.x)} and
where each element is an integer seed that can be assigned to \code{.Random.seed}.
Use this alternative with caution. \emph{Note that \code{as.list(seq_along(.x))} is
not a valid set of such \code{.Random.seed} values.}

In all cases but \code{seed = FALSE}, after a furrr function returns, the RNG
state of the calling R process is guaranteed to be "forwarded one step" from
the RNG state before the call. This is true regardless of the future
strategy / scheduling used. This is done in order to guarantee that an R
script calling \code{future_map()} multiple times should be numerically
reproducible given the same initial seed.

Note that you cannot expect identical results between \code{map()} and
\code{future_map()} when using a \code{.f} that calls functions that generate random
numbers, even when calling \code{set.seed()} ahead of time. For one thing, the
default random number generation algorithm used by R during sequential
processing is Mersenne-Twister, different from the L'Ecuyer-CMRG seeds used
by furrr. But even aligning the \code{\link[=RNGkind]{RNGkind()}} would not be enough. \code{map()}
itself would have to change to use the same parallel compatible RNG strategy
as \code{future_map()} (pre-generating the seeds, and setting them before each
\code{.f} invocation). At the end of the day, you have to accept that the
following will produce different sequences of random numbers, but both are
statistically sound:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{set.seed(42)
purrr::map(1:10, ~ rnorm(1))

set.seed(42)
furrr::future_map(1:10, ~ rnorm(1), .options = furrr_options(seed = TRUE))
}\if{html}{\out{</div>}}

But importantly, the \code{furrr::future_map()} example will always produce the
same sequence of random numbers, regardless of the \code{plan()} you choose:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{plan(sequential)
set.seed(42)
furrr::future_map(1:10, ~ rnorm(1), .options = furrr_options(seed = TRUE))

plan(multisession, workers = 2)
set.seed(42)
furrr::future_map(1:10, ~ rnorm(1), .options = furrr_options(seed = TRUE))

plan(cluster, workers = workers)
set.seed(42)
furrr::future_map(1:10, ~ rnorm(1), .options = furrr_options(seed = TRUE))
}\if{html}{\out{</div>}}
}

\examples{
furrr_options()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-map2.R, R/future-pmap.R, R/future-walk.R
\name{future_map2}
\alias{future_map2}
\alias{future_map2_chr}
\alias{future_map2_dbl}
\alias{future_map2_int}
\alias{future_map2_lgl}
\alias{future_map2_vec}
\alias{future_map2_dfr}
\alias{future_map2_dfc}
\alias{future_pmap}
\alias{future_pmap_chr}
\alias{future_pmap_dbl}
\alias{future_pmap_int}
\alias{future_pmap_lgl}
\alias{future_pmap_vec}
\alias{future_pmap_dfr}
\alias{future_pmap_dfc}
\alias{future_walk2}
\alias{future_pwalk}
\title{Map over multiple inputs simultaneously via futures}
\usage{
future_map2(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map2_chr(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map2_dbl(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map2_int(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map2_lgl(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map2_vec(
  .x,
  .y,
  .f,
  ...,
  .ptype = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map2_dfr(
  .x,
  .y,
  .f,
  ...,
  .id = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map2_dfc(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_pmap(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_pmap_chr(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_pmap_dbl(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_pmap_int(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_pmap_lgl(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_pmap_vec(
  .l,
  .f,
  ...,
  .ptype = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_pmap_dfr(
  .l,
  .f,
  ...,
  .id = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_pmap_dfc(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_walk2(
  .x,
  .y,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_pwalk(
  .l,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)
}
\arguments{
\item{.x, .y}{A pair of vectors, usually the same length. If not, a vector
of length 1 will be recycled to the length of the other.}

\item{.f}{A function, specified in one of the following ways:
\itemize{
\item A named function.
\item An anonymous function, e.g. \verb{\\(x, y) x + y} or \code{function(x, y) x + y}.
\item A formula, e.g. \code{~ .x + .y}. Use \code{.x} to refer to the current
element of \code{x} and \code{.y} to refer to the current element of \code{y}.
No longer recommended.
}}

\item{...}{Additional arguments passed on to the mapped function.

We now generally recommend against using \code{...} to pass additional
(constant) arguments to \code{.f}. Instead use a shorthand anonymous function:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{# Instead of
x |> future_map(f, 1, 2, collapse = ",")
# do:
x |> future_map(\\(x) f(x, 1, 2, collapse = ","))
}\if{html}{\out{</div>}}

This makes it easier to understand which arguments belong to which
function and will tend to yield better error messages.}

\item{.options}{The \code{future} specific options to use with the workers. This
must be the result from a call to \code{\link[=furrr_options]{furrr_options()}}.}

\item{.env_globals}{The environment to look for globals required by \code{.x} and
\code{...}. Globals required by \code{.f} are looked up in the function environment
of \code{.f}.}

\item{.progress}{A single logical. Should a progress bar be displayed?
Only works with multisession, multicore, and multiprocess futures. Note
that if a multicore/multisession future falls back to sequential, then
a progress bar will not be displayed.

\strong{Warning:} The \code{.progress} argument will be deprecated and removed
in a future version of furrr in favor of using the more robust
\href{https://CRAN.R-project.org/package=progressr}{progressr}
package.}

\item{.ptype}{If \code{NULL}, the default, the output type is the common type of
the elements of the result. Otherwise, supply a "prototype" giving the
desired type of output.}

\item{.id}{Either a string or \code{NULL}. If a string, the output will contain
a variable with that name, storing either the name (if \code{.x} is named) or
the index (if \code{.x} is unnamed) of the input. If \code{NULL}, the default, no
variable will be created.

Only applies to \verb{_dfr} variant.}

\item{.l}{A list of vectors. The length of \code{.l} determines the number of
arguments that \code{.f} will be called with. Arguments will be supply by
position if unnamed, and by name if named.

Vectors of length 1 will be recycled to any length; all other elements
must be have the same length.

A data frame is an important special case of \code{.l}. It will cause \code{.f}
to be called once for each row.}
}
\value{
An atomic vector, list, or data frame, depending on the suffix.
Atomic vectors and lists will be named if \code{.x} or the first element of \code{.l}
is named.

If all input is length 0, the output will be length 0. If any input is
length 1, it will be recycled to the length of the longest.
}
\description{
These functions work the same as \code{\link[purrr:map2]{purrr::map2()}} and its variants,
but allow you to map in parallel. Note that "parallel" as described in purrr
is just saying that you are working with multiple inputs, and parallel in
this case means that you can work on multiple inputs and process them all in
parallel as well.
}
\examples{
\donttest{plan(multisession, workers = 2)}

x <- list(1, 10, 100)
y <- list(1, 2, 3)
z <- list(5, 50, 500)

future_map2(x, y, ~ .x + .y)

# Split into pieces, fit model to each piece, then predict
by_cyl <- split(mtcars, mtcars$cyl)
mods <- future_map(by_cyl, ~ lm(mpg ~ wt, data = .))
future_map2(mods, by_cyl, predict)

future_pmap(list(x, y, z), sum)

# Matching arguments by position
future_pmap(list(x, y, z), function(a, b ,c) a / (b + c))

# Vectorizing a function over multiple arguments
df <- data.frame(
  x = c("apple", "banana", "cherry"),
  pattern = c("p", "n", "h"),
  replacement = c("x", "f", "q"),
  stringsAsFactors = FALSE
)

future_pmap(df, gsub)
future_pmap_chr(df, gsub)

\dontshow{
# Close open connections for R CMD Check
if (!inherits(plan(), "sequential")) plan(sequential)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-imap.R, R/future-walk.R
\name{future_imap}
\alias{future_imap}
\alias{future_imap_chr}
\alias{future_imap_dbl}
\alias{future_imap_int}
\alias{future_imap_lgl}
\alias{future_imap_vec}
\alias{future_imap_dfr}
\alias{future_imap_dfc}
\alias{future_iwalk}
\title{Apply a function to each element of a vector, and its index via futures}
\usage{
future_imap(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_imap_chr(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_imap_dbl(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_imap_int(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_imap_lgl(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_imap_vec(
  .x,
  .f,
  ...,
  .ptype = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_imap_dfr(
  .x,
  .f,
  ...,
  .id = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_imap_dfc(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_iwalk(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)
}
\arguments{
\item{.x}{A list or atomic vector.}

\item{.f}{A function, specified in one of the following ways:
\itemize{
\item A named function, e.g. \code{paste}.
\item An anonymous function, e.g. \verb{\\(x, idx) x + idx} or
\code{function(x, idx) x + idx}.
\item A formula, e.g. \code{~ .x + .y}. Use \code{.x} to refer to the current element and
\code{.y} to refer to the current index. No longer recommended.
}}

\item{...}{Additional arguments passed on to the mapped function.

We now generally recommend against using \code{...} to pass additional
(constant) arguments to \code{.f}. Instead use a shorthand anonymous function:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{# Instead of
x |> future_map(f, 1, 2, collapse = ",")
# do:
x |> future_map(\\(x) f(x, 1, 2, collapse = ","))
}\if{html}{\out{</div>}}

This makes it easier to understand which arguments belong to which
function and will tend to yield better error messages.}

\item{.options}{The \code{future} specific options to use with the workers. This
must be the result from a call to \code{\link[=furrr_options]{furrr_options()}}.}

\item{.env_globals}{The environment to look for globals required by \code{.x} and
\code{...}. Globals required by \code{.f} are looked up in the function environment
of \code{.f}.}

\item{.progress}{A single logical. Should a progress bar be displayed?
Only works with multisession, multicore, and multiprocess futures. Note
that if a multicore/multisession future falls back to sequential, then
a progress bar will not be displayed.

\strong{Warning:} The \code{.progress} argument will be deprecated and removed
in a future version of furrr in favor of using the more robust
\href{https://CRAN.R-project.org/package=progressr}{progressr}
package.}

\item{.ptype}{If \code{NULL}, the default, the output type is the common type of
the elements of the result. Otherwise, supply a "prototype" giving the
desired type of output.}

\item{.id}{Either a string or \code{NULL}. If a string, the output will contain
a variable with that name, storing either the name (if \code{.x} is named) or
the index (if \code{.x} is unnamed) of the input. If \code{NULL}, the default, no
variable will be created.

Only applies to \verb{_dfr} variant.}
}
\value{
A vector the same length as .x.
}
\description{
These functions work the same as \code{\link[purrr:imap]{purrr::imap()}} functions,
but allow you to map in parallel.
}
\examples{
\donttest{plan(multisession, workers = 2)}

future_imap_chr(sample(10), ~ paste0(.y, ": ", .x))

\dontshow{
# Close open connections for R CMD Check
if (!inherits(plan(), "sequential")) plan(sequential)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-map.R
\name{future_map_if}
\alias{future_map_if}
\alias{future_map_at}
\title{Apply a function to each element of a vector conditionally via futures}
\usage{
future_map_if(
  .x,
  .p,
  .f,
  ...,
  .else = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_map_at(
  .x,
  .at,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)
}
\arguments{
\item{.x}{A list or atomic vector.}

\item{.p}{A single predicate function, a formula describing such a
predicate function, or a logical vector of the same length as \code{.x}.
Alternatively, if the elements of \code{.x} are themselves lists of
objects, a string indicating the name of a logical element in the
inner lists. Only those elements where \code{.p} evaluates to
\code{TRUE} will be modified.}

\item{.f}{A function, specified in one of the following ways:
\itemize{
\item A named function, e.g. \code{mean}.
\item An anonymous function, e.g. \verb{\\(x) x + 1} or \code{function(x) x + 1}.
\item A formula, e.g. \code{~ .x + 1}. Use \code{.x} to refer to the first
argument. No longer recommended.
\item A string, integer, or list, e.g. \code{"idx"}, \code{1}, or \code{list("idx", 1)} which
are shorthand for \verb{\\(x) pluck(x, "idx")}, \verb{\\(x) pluck(x, 1)}, and
\verb{\\(x) pluck(x, "idx", 1)} respectively. Optionally supply \code{.default} to
set a default value if the indexed element is \code{NULL} or does not exist.
}}

\item{...}{Additional arguments passed on to the mapped function.

We now generally recommend against using \code{...} to pass additional
(constant) arguments to \code{.f}. Instead use a shorthand anonymous function:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{# Instead of
x |> future_map(f, 1, 2, collapse = ",")
# do:
x |> future_map(\\(x) f(x, 1, 2, collapse = ","))
}\if{html}{\out{</div>}}

This makes it easier to understand which arguments belong to which
function and will tend to yield better error messages.}

\item{.else}{A function applied to elements of \code{.x} for which \code{.p}
returns \code{FALSE}.}

\item{.options}{The \code{future} specific options to use with the workers. This
must be the result from a call to \code{\link[=furrr_options]{furrr_options()}}.}

\item{.env_globals}{The environment to look for globals required by \code{.x} and
\code{...}. Globals required by \code{.f} are looked up in the function environment
of \code{.f}.}

\item{.progress}{A single logical. Should a progress bar be displayed?
Only works with multisession, multicore, and multiprocess futures. Note
that if a multicore/multisession future falls back to sequential, then
a progress bar will not be displayed.

\strong{Warning:} The \code{.progress} argument will be deprecated and removed
in a future version of furrr in favor of using the more robust
\href{https://CRAN.R-project.org/package=progressr}{progressr}
package.}

\item{.at}{A logical, integer, or character vector giving the elements
to select. Alternatively, a function that takes a vector of names,
and returns a logical, integer, or character vector of elements to select.

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}: if the tidyselect package is
installed, you can use \code{vars()} and tidyselect helpers to select
elements.}
}
\value{
Both functions return a list the same length as \code{.x} with the
elements conditionally transformed.
}
\description{
These functions work the same as \code{\link[purrr:map_if]{purrr::map_if()}} and
\code{\link[purrr:map_if]{purrr::map_at()}}, but allow you to run them in parallel.
}
\examples{
\donttest{plan(multisession, workers = 2)}

# Modify the even elements
future_map_if(1:5, ~.x \%\% 2 == 0L, ~ -1)

future_map_at(1:5, c(1, 5), ~ -1)
\dontshow{
# Close open connections for R CMD Check
if (!inherits(plan(), "sequential")) plan(sequential)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-modify.R
\name{future_modify}
\alias{future_modify}
\alias{future_modify_at}
\alias{future_modify_if}
\title{Modify elements selectively via futures}
\usage{
future_modify(
  .x,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_modify_at(
  .x,
  .at,
  .f,
  ...,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)

future_modify_if(
  .x,
  .p,
  .f,
  ...,
  .else = NULL,
  .options = furrr_options(),
  .env_globals = parent.frame(),
  .progress = FALSE
)
}
\arguments{
\item{.x}{A vector.}

\item{.f}{A function specified in the same way as the corresponding map
function.}

\item{...}{Additional arguments passed on to the mapped function.

We now generally recommend against using \code{...} to pass additional
(constant) arguments to \code{.f}. Instead use a shorthand anonymous function:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{# Instead of
x |> future_map(f, 1, 2, collapse = ",")
# do:
x |> future_map(\\(x) f(x, 1, 2, collapse = ","))
}\if{html}{\out{</div>}}

This makes it easier to understand which arguments belong to which
function and will tend to yield better error messages.}

\item{.options}{The \code{future} specific options to use with the workers. This
must be the result from a call to \code{\link[=furrr_options]{furrr_options()}}.}

\item{.env_globals}{The environment to look for globals required by \code{.x} and
\code{...}. Globals required by \code{.f} are looked up in the function environment
of \code{.f}.}

\item{.progress}{A single logical. Should a progress bar be displayed?
Only works with multisession, multicore, and multiprocess futures. Note
that if a multicore/multisession future falls back to sequential, then
a progress bar will not be displayed.

\strong{Warning:} The \code{.progress} argument will be deprecated and removed
in a future version of furrr in favor of using the more robust
\href{https://CRAN.R-project.org/package=progressr}{progressr}
package.}

\item{.at}{A logical, integer, or character vector giving the elements
to select. Alternatively, a function that takes a vector of names,
and returns a logical, integer, or character vector of elements to select.

\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}: if the tidyselect package is
installed, you can use \code{vars()} and tidyselect helpers to select
elements.}

\item{.p}{A single predicate function, a formula describing such a
predicate function, or a logical vector of the same length as \code{.x}.
Alternatively, if the elements of \code{.x} are themselves lists of
objects, a string indicating the name of a logical element in the
inner lists. Only those elements where \code{.p} evaluates to
\code{TRUE} will be modified.}

\item{.else}{A function applied to elements of \code{.x} for which \code{.p}
returns \code{FALSE}.}
}
\value{
An object the same class as \code{.x}
}
\description{
These functions work the same as \code{\link[purrr:modify]{purrr::modify()}} functions, but
allow you to modify in parallel.
}
\details{
From purrr:

Since the transformation can alter the structure of the input;
it's your responsibility to ensure that the transformation produces a valid
output. For example, if you're modifying a data frame, \code{.f} must preserve the
length of the input.
}
\examples{
\donttest{plan(multisession, workers = 2)}

# Convert each col to character, in parallel
future_modify(mtcars, as.character)

iris |>
 future_modify_if(is.factor, as.character) |>
 str()

mtcars |>
  future_modify_at(c(1, 4, 5), as.character) |>
  str()

\dontshow{
# Close open connections for R CMD Check
if (!inherits(plan(), "sequential")) plan(sequential)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/furrr-package.R
\docType{package}
\name{furrr-package}
\alias{furrr}
\alias{furrr-package}
\title{furrr: Apply Mapping Functions in Parallel using Futures}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Implementations of the family of map() functions from 'purrr' that can be resolved using any 'future'-supported backend, e.g. parallel on the local machine or distributed on a compute cluster.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/futureverse/furrr}
  \item \url{https://furrr.futureverse.org/}
  \item Report bugs at \url{https://github.com/futureverse/furrr/issues}
}

}
\author{
\strong{Maintainer}: Davis Vaughan \email{davis@posit.co} (\href{https://orcid.org/0000-0003-4777-038X}{ORCID})

Authors:
\itemize{
  \item Henrik Bengtsson \email{henrikb@braju.com} (\href{https://orcid.org/0000-0002-7579-5165}{ORCID})
  \item Matt Dancho \email{mdancho@business-science.io}
}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}
